<?php

namespace Drupal\aws_cloud\EventSubscriber;

use Drupal\aws_cloud\Service\Ec2\Ec2ServiceException;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\cloud\Event\CloudEntityEventType;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\cloud\Event\CloudEntityEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Event subscriber to handle updating duplicate AWS cloud service providers.
 */
class AwsEntityUpdateSubscriber implements EventSubscriberInterface {

  use CloudContentEntityTrait;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  private $cloudConfigPluginManager;

  /**
   * The Ec2Service object.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * AwsEntityUpdateSubscriber constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager interface.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud config plugin manager.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The EC2 Service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, CloudConfigPluginManagerInterface $cloud_config_plugin_manager, Ec2ServiceInterface $ec2_service) {
    $this->entityTypeManager = $entity_type_manager;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->ec2Service = $ec2_service;
  }

  /**
   * Delete AWS launch template.
   *
   * @param \Drupal\cloud\Event\CloudEntityEvent $event
   *   The event object.
   *
   * @return bool
   *   TRUE if successful, otherwise return FALSE.
   *
   * @throws \Drupal\aws_cloud\Service\Ec2\Ec2ServiceException
   *   Throws Ec2ServiceException if an error occurs.
   */
  public function deleteLaunchTemplate(CloudEntityEvent $event): bool {

    $entity = $event->getEntity();

    if (empty($entity)
      || $entity->getEntityTypeId() !== 'cloud_server_template'
      || $entity->bundle() !== 'aws_cloud') {
      return FALSE;
    }

    try {

      $params = [
        'LaunchTemplateName' => $entity->getName(),
      ];
      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $this->ec2Service->deleteLaunchTemplate($params);
    }
    catch (\Exception $e) {
      throw new Ec2ServiceException($this->t('Failed to delete AWS launch template.'));
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events = [
      // When an AWS launch template is deleted.
      CloudEntityEventType::LAUNCH_TEMPLATE_DELETE => 'deleteLaunchTemplate',
    ];

    return $events;
  }

}
