<?php

namespace Drupal\Tests\aws_cloud\Functional\cloud\config;

use Drupal\cloud\Entity\CloudContentEntityBase;
use Drupal\Tests\aws_cloud\Traits\AwsCloudTestFormDataTrait;
use Drupal\Tests\aws_cloud\Traits\AwsCloudTestMockTrait;
use Drupal\Tests\cloud\Functional\cloud\config\CloudConfigTestBase;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;

/**
 * Tests cloud service provider (CloudConfig).
 *
 * @group Cloud
 */
class AwsCloudConfigPermissionTest extends CloudConfigTestBase {

  use AwsCloudTestFormDataTrait;
  use AwsCloudTestMockTrait;

  /**
   * AWS_CLOUD_CONFIG_REPEAT_COUNT.
   *
   * @var int
   */
  public const AWS_CLOUD_CONFIG_REPEAT_COUNT = 2;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'cloud',
    'aws_cloud',
  ];

  /**
   * {@inheritdoc}
   */
  protected function getPermissions(): array {
    return array_merge([
      'administer cloud service providers',
      'add cloud service providers',
      'edit cloud service providers',
      'edit own cloud service providers',
      'delete cloud service providers',
      'delete own cloud service providers',
      'view published cloud service providers',
      'view own published cloud service providers',
      'access dashboard',
      'view cloud service provider admin list',
      'list cloud server template',
      'administer aws_cloud',
      'view aws cloud instance type prices',
      'list aws cloud instance',
    ], $this->getOptionalPermissions());
  }

  /**
   * Get optional permissions. In case of lacking, resulting in warning.
   *
   * @return string[]
   *   Optional permissions.
   */
  private function getOptionalPermissions(): array {
    return [
      'add aws cloud key pair',
      'add aws cloud network interface',
      'add aws cloud security group',
      'add aws cloud service provider',
      'add aws cloud subnet',
      'add aws cloud vpc',
      'delete own aws cloud security group',
      'edit own aws cloud instance',
      'edit own aws cloud security group',
      'list aws cloud carrier gateway',
      'list aws cloud elastic ip',
      'list aws cloud image',
      'list aws cloud internet gateway',
      'list aws cloud key pair',
      'list aws cloud network interface',
      'list aws cloud security group',
      'list aws cloud snapshot',
      'list aws cloud subnet',
      'list aws cloud transit gateway',
      'list aws cloud volume',
      'list aws cloud vpc peering connection',
      'list aws cloud vpc',
      'view own aws cloud instance',
    ];
  }

  /**
   * Get permissions to validate.
   *
   * @FIXEME no duplicate definition.
   *
   * @return array
   *   An associative array of permissions.
   */
  private function getPermissionsToValidate(): array {
    return [
      'mandatory' => [
        // EC2 APIs.
        // Excluding 'DescribeInstances' as this check is included in
        // aws_cloud_form_cloud_config_aws_cloud_describe_instances_validate().
        'DescribeAddresses' => 'list aws cloud elastic ip',
        'DescribeAvailabilityZones' => 'list aws cloud instance',
        'DescribeCarrierGateways' => 'list aws cloud carrier gateway',
        'DescribeFlowLogs' => 'list aws cloud vpc',
        'DescribeIamInstanceProfileAssociations' => 'list aws cloud instance',
        'DescribeImageAttribute' => 'list aws cloud image',
        'DescribeImages' => 'list aws cloud image',
        'DescribeInstanceAttribute' => 'list aws cloud instance',
        'DescribeInternetGateways' => 'list aws cloud internet gateway',
        'DescribeKeyPairs' => 'list aws cloud key pair',
        'DescribeNetworkInterfaces' => 'list aws cloud network interface',
        'DescribeSecurityGroups' => 'list aws cloud security group',
        'DescribeSnapshots' => 'list aws cloud snapshot',
        'DescribeSubnets' => 'list aws cloud subnet',
        'DescribeTransitGateways' => 'list aws cloud transit gateway',
        'DescribeVolumes' => 'list aws cloud volume',
        'DescribeVpcPeeringConnections' => 'list aws cloud vpc peering connection',
        'DescribeVpcs' => 'list aws cloud vpc',
        'DescribeLaunchTemplates' => 'list cloud server template',
        'DescribeLaunchTemplateVersions' => 'list cloud server template',
        // IAM APIs.
        'ListInstanceProfiles' => 'add aws cloud service provider',
      ],
      'optional' => [
        // EC2 APIs.
        'AuthorizeSecurityGroupEgress' => 'edit own aws cloud security group',
        'AuthorizeSecurityGroupIngress' => 'edit own aws cloud security group',
        'CreateKeyPair' => 'add aws cloud key pair',
        'CreateNetworkInterface' => 'add aws cloud network interface',
        'CreateSecurityGroup' => 'add aws cloud security group',
        'CreateSubnet' => 'add aws cloud subnet',
        'CreateTags' => 'edit own aws cloud instance',
        'CreateVpc' => 'add aws cloud vpc',
        'GetConsoleOutput' => 'view own aws cloud instance',
        'RevokeSecurityGroupEgress' => 'delete own aws cloud security group',
        'RevokeSecurityGroupIngress' => 'delete own aws cloud security group',
        // Cloud Watch APIs.
        'GetMetricData' => 'view own aws cloud instance',
      ],
    ];
  }

  /**
   * Get permissions that those validation depends on other permission.
   *
   * @return array
   *   An associative array.
   */
  private function getPermissionsDependingOn(): array {
    return [
      'DescribeSecurityGroups' => [
        'AuthorizeSecurityGroupEgress',
        'AuthorizeSecurityGroupIngress',
        'CreateTags',
        'RevokeSecurityGroupEgress',
        'RevokeSecurityGroupIngress',
      ],
    ];
  }

  /**
   * Flatten permissions.
   *
   * @return array
   *   Permissions.
   */
  private function flattenPermissions(): array {
    return array_keys(array_merge([], ...array_values($this->getPermissionsToValidate())));
  }

  /**
   * Create cloud context.
   *
   * @param string $bundle
   *   The CloudConfig Bundle Type ('aws_cloud').
   *
   * @return \Drupal\cloud\Entity\CloudConfig
   *   The cloud service provider (CloudConfig) entity.
   */
  protected function createCloudContext($bundle = __CLASS__): CloudContentEntityBase {
    return parent::createCloudContext($this->getModuleName($bundle));
  }

  /**
   * Set up test.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  protected function setUp(): void {

    parent::setUp();

    $this->init(__CLASS__, $this);
    $this->initMockInstanceTypes();

    // Delete the existing $this->cloudContext since we test a CloudConfig
    // entities multiple deletion for themselves.
    if (!empty($this->cloudConfig)) {
      $this->cloudConfig->delete();
    }

    $this->setUpAddOperation();
  }

  /**
   * {@inheritdoc}
   */
  protected function getMockDataTemplateVars(): array {
    return [
      'group_id'    => 'sg-' . $this->getRandomId(),
      'error_code'  => 'DryRunOperation',
    ];
  }

  /**
   * Set up before testing Create operation.
   *
   * @throws \Exception
   */
  private function setUpAddOperation(): void {

    // List AWS Cloud service providers for Amazon EC2.
    $this->drupalGet('/admin/structure/cloud_config');
    $this->assertNoErrorMessage();

    // Add random VPC information (Server-side).
    $vpcs = $this->createVpcsRandomTestFormData();
    $subnets = $this->createSubnetsRandomTestFormData();
    $this->updateVpcsAndSubnetsMockData($vpcs, $subnets);

    // Add random Image information (Server-side).
    $images = $this->createImagesRandomTestFormData();
    $this->updateImagesMockData($images);
  }

  /**
   * Set up detailed data before save operation.
   *
   * @param int $repeat_index
   *   Test repeat index.
   * @param array $add
   *   Cloud config data to add.
   * @param int $target_region_index
   *   Target region index to add the cloud config.
   *
   * @throws \Exception
   */
  private function setUpDetailsBeforeSave(int $repeat_index, array &$add, int $target_region_index): void {

    // Get all regions.
    $regions = \Drupal::service('aws_cloud.ec2')->getRegions();
    // We need to specify the 'view <CLOUD_CONTEXT>' permission.
    $region_index = 0;

    foreach ($regions ?: [] as $region_name => $region_display_name) {
      if ($region_index === $target_region_index) {
        $add[$repeat_index]['regions[]'][] = $region_name;
      }
      $this->grantPermissions(
        Role::load(
          RoleInterface::AUTHENTICATED_ID),
        [
          'view ' . aws_cloud_form_cloud_config_aws_cloud_add_form_create_cloud_context(
            $add[$repeat_index]['name[0][value]'],
            $region_name),
        ]
      );

      $region_index++;
    }

    // These variables that are sent from a browser are specified in the form
    // in the browser, therefore we don't use the auto-generated test values
    // by createCloudConfigTestFormData.
    unset(
      $add[$repeat_index]['field_region'],
      $add[$repeat_index]['field_api_endpoint_uri'],
      $add[$repeat_index]['field_image_upload_url']
    );

    // Enable IAM permission check.
    $add[$repeat_index]['field_check_iam_permissions[value]'] = 1;
    // Set default Security Group for IAM permission check.
    $this->addSecurityGroupMockData('default', $this->getRandomSecurityGroupDescription(), $this->getRandomId());

  }

  /**
   * Assert warning and error messages based on unset IAM permissions.
   *
   * @param string $unset_mandatory_perm
   *   A mandatory IAM permission to unset.
   * @param string $unset_optional_perm
   *   A optional IAM permission to unset.
   */
  private function assertWarningAndErrorsBasedOnIamPermissions(string $unset_mandatory_perm, string $unset_optional_perm): void {

    // Asserting an error message regardless of randomly selected permissions.
    $t_args = [
      '@to_fix' => 'credentials',
      '@unauthorized_apis' => $unset_mandatory_perm,
    ];
    $this->assertSession()
      ->pageTextContains(strip_tags($this->t('Please add the following IAM permission(s) to your @to_fix: @unauthorized_apis', $t_args)));

    if (!empty($this->getPermissionsDependingOn()[$unset_mandatory_perm])
      && in_array($unset_optional_perm, $this->getPermissionsDependingOn()[$unset_mandatory_perm], TRUE)) {
      // If the removed optional permission depends on the removed mandatory
      // permission, a warning message is to be suppressed. Only error
      // message appears.
      $this->assertErrorMessage();
      return;
    }

    $t_args = [
      '@to_fix' => 'credentials',
      '@unauthorized_apis' => $unset_optional_perm,
    ];
    $this->assertSession()
      ->pageTextContains(strip_tags($this->t('You might want to add the following IAM permission(s) to your @to_fix: @unauthorized_apis', $t_args)));
    $this->assertWarningAndErrorMessage();

  }

  /**
   * Tests Create operation for cloud service provider expecting errors.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Exception
   */
  public function testAddCloudServiceProviderWithErrors(): void {

    // Add a new Config information.
    $add = $this->createCloudConfigTestFormData(self::AWS_CLOUD_CONFIG_REPEAT_COUNT);
    // Get all regions.
    $regions = \Drupal::service('aws_cloud.ec2')->getRegions();
    $target_region_index = random_int(0, count($regions) - 1);

    for ($i = 0; $i < self::AWS_CLOUD_CONFIG_REPEAT_COUNT; $i++) {
      $this->setUpDetailsBeforeSave($i, $add, $target_region_index);

      // Set all required IAM permissions.
      $this->setAwsCloudConfigPermissionMockData($this->flattenPermissions());
      // For error-expected testing, unset randomly-picked IAM permissions.
      $unset_mandatory_perm = array_rand($this->getPermissionsToValidate()['mandatory']);
      $unset_optional_perm = array_rand($this->getPermissionsToValidate()['optional']);

      $errorsAndWarnings = random_int(0, 1) === 1;
      $this->updateAwsCloudConfigPermissionMockData([
        $unset_mandatory_perm,
        $unset_optional_perm,
      ], $errorsAndWarnings);

      // Make sure no error message on loading an add form.
      $this->drupalGet('/admin/structure/cloud_config/add');
      $this->assertNoErrorMessage();

      $this->submitForm(
        $add[$i],
        $this->t('Save'));

      $errorsAndWarnings
        ? $this->assertWarningAndErrorsBasedOnIamPermissions($unset_mandatory_perm, $unset_optional_perm)
        : $this->assertErrorMessage();
      // Assert no additional errors.
      $this->assertSession()->pageTextNotContains($this->t('An error occurred'));
      $this->assertSession()->pageTextNotContains($this->t('An illegal choice has been detected. Please contact the site administrator.'));
    }
  }

  /**
   * Tests Update operation for cloud service provider expecting errors.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Exception
   */
  public function testEditCloudServiceProviderWithErrors(): void {

    // Add a new Config information only once regardless of
    // self::AWS_CLOUD_CONFIG_REPEAT_COUNT.
    $add = $this->createCloudConfigTestFormData(1);
    // Get all regions.
    $regions = \Drupal::service('aws_cloud.ec2')->getRegions();
    $target_region_index = random_int(0, count($regions) - 1);
    $target_region_display_name = $regions[array_keys($regions)[$target_region_index]];

    $repeat_index = 0;
    $this->setUpDetailsBeforeSave($repeat_index, $add, $target_region_index);
    $label = "{$add[$repeat_index]['name[0][value]']} {$target_region_display_name}";

    // To avoid DryRunOperation AwdException, empty the error code.
    $this->setAwsCloudConfigPermissionMockData($this->flattenPermissions(), '');
    // Make sure no error message on showing an add form.
    $this->drupalGet('/admin/structure/cloud_config/add');
    $this->assertNoErrorMessage();

    $this->submitForm(
      $add[$repeat_index],
      $this->t('Save'));

    // Expect no error on add test and proceed with edit testing.
    $this->assertErrorMessage();
    $this->assertGoogleAppErrorOnly();
    $this->assertSession()->pageTextContains($this->t('Status message'));
    $this->assertSession()
      ->pageTextContains($this->t('Creating cloud service provider was performed successfully.'));
    $this->assertSession()->pageTextContains($label);

    // Make sure listing for '/admin/config/services/cloud/aws_cloud'.
    $this->drupalGet('/admin/config/services/cloud/aws_cloud');
    $this->assertNoErrorMessage();
    $this->assertSession()->linkExists($label);

    // Edit Config case.
    $edit = $this->createCloudConfigTestFormData(self::AWS_CLOUD_CONFIG_REPEAT_COUNT);

    // This is CloudConfig test case, so we don't require default
    // $this->cloudContext, which has been already deleted in this setUp().
    // The entity number of $this->cloudContext was '1'.  Therefore the entity
    // number starts from '2', not '1', here.
    for ($i = 0, $num = 2; $i < self::AWS_CLOUD_CONFIG_REPEAT_COUNT; $i++) {

      // Set all required IAM permissions.
      $this->setAwsCloudConfigPermissionMockData($this->flattenPermissions());
      // For error-expected testing, unset randomly-picked IAM permissions.
      $unset_mandatory_perm = array_rand($this->getPermissionsToValidate()['mandatory']);
      $unset_optional_perm = array_rand($this->getPermissionsToValidate()['optional']);

      $errorsAndWarnings = random_int(0, 1) === 1;
      $this->updateAwsCloudConfigPermissionMockData([
        $unset_mandatory_perm,
        $unset_optional_perm,
      ], $errorsAndWarnings);

      // Make sure no error message on showing an edit form.
      $this->drupalGet("/admin/structure/cloud_config/${num}/edit");
      $this->assertNoErrorMessage();

      $this->submitForm(
        $edit[$i],
        $this->t('Save'));

      $errorsAndWarnings
        ? $this->assertWarningAndErrorsBasedOnIamPermissions($unset_mandatory_perm, $unset_optional_perm)
        : $this->assertErrorMessage();
      // Assert no additional errors.
      $this->assertSession()->pageTextNotContains($this->t('An error occurred'));
      $this->assertSession()->pageTextNotContains($this->t('An illegal choice has been detected. Please contact the site administrator.'));
    }
  }

}
