<?php

namespace Drupal\k8s\Form;

use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Entity\EntityMalformedException;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Form\FormStateInterface;
use Drupal\k8s\Entity\K8sEntityBase;
use Drupal\k8s\Service\K8sServiceException;

/**
 * Form controller for the entity create form.
 *
 * @ingroup k8s
 */
class K8sCreateForm extends K8sContentForm {

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $cloud_context = '') {
    $form = parent::buildForm($form, $form_state);
    $this->k8sService->setCloudContext($cloud_context);

    $name_underscore = $this->getShortEntityTypeNameUnderscore($this->entity);
    $name_whitespace = $this->getShortEntityTypeNameWhitespace($this->entity);

    $weight = -50;

    $form[$name_underscore] = [
      '#type' => 'details',
      '#title' => $name_whitespace,
      '#open' => TRUE,
      '#weight' => $weight++,
    ];

    $options = [];
    try {
      $options = $this->getNamespaceOptions();
    }
    catch (\Exception $e) {
      $this->k8sService->handleError($e, $cloud_context, $this->entity);
    }

    if (!empty($options)
    && method_exists($this->entity, 'getNamespace')) {

      $form[$name_underscore]['namespace'] = [
        '#type'          => 'select',
        '#title'         => $this->t('Namespace'),
        '#options'       => $options,
        '#default_value' => 'default',
        '#required'      => TRUE,
      ];
    }

    $form[$name_underscore]['detail'] = $form['detail'];
    unset($form['detail']);

    $this->addOthersFieldset($form, $weight++, $cloud_context);

    $form['actions'] = $this->actions($form, $form_state, $cloud_context);

    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\k8s\Service\K8sServiceException
   *    Thrown when unable to create entity.
   */
  public function save(array $form, FormStateInterface $form_state) {
    $this->trimTextfields($form, $form_state);

    $entity = $this->entity;

    $cloud_context = $this->routeMatch->getParameter('cloud_context');
    $this->k8sService->setCloudContext($cloud_context);

    $this->saveYamlsOfEntityDetail($form_state);

    $form_state->setRedirect(
      "view.k8s_{$this->getShortEntityTypeNameUnderscore($entity)}.list",
      ['cloud_context' => $entity->getCloudContext()]
    );
  }

  /**
   * Update yaml information to entity.
   *
   * @param array $yaml
   *   The yaml.
   */
  protected function updateYamlToEntity(array $yaml): bool {
    $entity = $this->entity;

    $name_camel = !empty($yaml['kind']) ? $yaml['kind'] : $this->getShortEntityTypeNameCamel($entity);
    $name_underscore = self::getSnakeCase($name_camel);

    if (!empty($entity->getOwner())) {
      // Add owner uid to annotations.
      $yaml['metadata']['annotations'][K8sEntityBase::ANNOTATION_CREATED_BY_UID] = $entity->getOwner()->id();

      if ($name_underscore === 'deployment') {
        // Add owner uid to pod template.
        $yaml['spec']['template']['metadata']['annotations'][K8sEntityBase::ANNOTATION_CREATED_BY_UID] = $entity->getOwner()->id();
      }
    }

    try {
      $method_name = "create{$name_camel}";
      $result = method_exists($this->entity, 'getNamespace')
        ? $this->k8sService->$method_name($entity->getNamespace(), $yaml)
        : $this->k8sService->$method_name($yaml);

      $result = array_merge($yaml, $result);
      $entity = $this->k8sService->createEntityFromYaml($result);

      $entity->setName($result['metadata']['name']);
      if (method_exists($entity, 'setCreationYaml')) {
        $entity->setCreationYaml(Yaml::encode($yaml));
      }
      if (method_exists($entity, 'setNamespace')) {
        $entity->setNamespace($result['metadata']['namespace']);
      }

      $entity->save();

      // Update the entity.
      $name_plural_camel = $this->getShortEntityTypeNamePluralCamel($entity);
      $method_name = "update${name_plural_camel}";
      $this->k8sService->$method_name([
        'metadata.name' => $entity->getName(),
      ], FALSE);

      $this->processOperationStatus($entity, 'created');

      $this->entity = $entity;
    }
    catch (K8sServiceException
    | EntityStorageException
    | EntityMalformedException $e) {
      $this->logger('k8s')->error($e->getMessage());

      try {
        $this->processOperationErrorStatus($entity, 'created');
      }
      catch (EntityMalformedException $e) {
        $this->handleException($e);
      }
      return FALSE;
    }

    return TRUE;
  }

}
