<?php

namespace Drupal\k8s\Plugin\cloud\server_template;

use Drupal\cloud\Entity\CloudServerTemplateInterface;
use Drupal\cloud\Plugin\cloud\CloudPluginBase;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Plugin\cloud\server_template\CloudServerTemplatePluginInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Service\EntityLinkRendererInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Component\Serialization\Yaml;
use Drupal\Component\Serialization\Exception\InvalidDataTypeException;
use Drupal\Component\Utility\Html;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\Element\EntityAutocomplete;
use Drupal\Core\Entity\EntityMalformedException;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\Exception\NotRegularDirectoryException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Template\TwigEnvironment;
use Drupal\k8s\Entity\K8sEntityBase;
use Drupal\k8s\Entity\K8sNamespace;
use Drupal\k8s\Entity\K8sSchedule;
use Drupal\k8s\Service\K8sServiceException;
use Drupal\k8s\Service\K8sServiceInterface;
use Drupal\k8s\Traits\K8sFormTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\k8s\Plugin\EntityReferenceSelection\LaunchTemplateUserSelection;
use Drupal\user\Entity\User;
use GuzzleHttp\Promise\Promise;
use Drupal\Core\Render\Markup;

/**
 * K8s Cloud launch template plugin.
 */
class K8sCloudServerTemplatePlugin extends CloudPluginBase implements CloudServerTemplatePluginInterface, ContainerFactoryPluginInterface {

  use CloudContentEntityTrait;
  use K8sFormTrait;

  public const CHECK_RETRY_COUNT = 5;

  public const CHECK_SLEEP_SECOND = 1;

  /**
   * The K8s Service.
   *
   * @var \Drupal\k8s\Service\K8sServiceInterface
   */
  protected $k8sService;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * Current login user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Entity link renderer object.
   *
   * @var \Drupal\cloud\Service\EntityLinkRendererInterface
   */
  protected $entityLinkRenderer;

  /**
   * File system object.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  protected $cloudService;

  /**
   * Twig.
   *
   * @var \Drupal\Core\Template\TwigEnvironment
   */
  protected $twig;

  /**
   * K8sCloudServerTemplatePlugin constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_service
   *   The uuid service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   Current login user.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\EntityLinkRendererInterface $entity_link_renderer
   *   The entity link render service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\k8s\Service\K8sServiceInterface $k8s_service
   *   The K8s Service.
   * @param \Drupal\Core\Template\TwigEnvironment $twig
   *   Twig.
   */
  public function __construct(array $configuration,
                              $plugin_id,
                              $plugin_definition,
                              EntityTypeManagerInterface $entity_type_manager,
                              UuidInterface $uuid_service,
                              AccountProxyInterface $current_user,
                              ConfigFactoryInterface $config_factory,
                              CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
                              EntityLinkRendererInterface $entity_link_renderer,
                              FileSystemInterface $file_system,
                              CloudServiceInterface $cloud_service,
                              K8sServiceInterface $k8s_service,
                              TwigEnvironment $twig) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->entityTypeManager = $entity_type_manager;
    $this->uuidService = $uuid_service;
    $this->currentUser = $current_user;
    $this->configFactory = $config_factory;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->entityLinkRenderer = $entity_link_renderer;
    $this->fileSystem = $file_system;
    $this->cloudService = $cloud_service;
    $this->k8sService = $k8s_service;
    $this->twig = $twig;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('uuid'),
      $container->get('current_user'),
      $container->get('config.factory'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('entity.link_renderer'),
      $container->get('file_system'),
      $container->get('cloud'),
      $container->get('k8s'),
      $container->get('twig')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityBundleName() {
    return $this->pluginDefinition['entity_bundle'];
  }

  /**
   * {@inheritdoc}
   */
  public function launch(CloudServerTemplateInterface $cloud_server_template, FormStateInterface $form_state = NULL) {
    $k8s_obj = $cloud_server_template->get('field_object')->value;
    $source_type = $cloud_server_template->get('field_source_type')->value;
    $delete_resources = $form_state->getValue('field_delete_all_resources');
    $launch_uid = $form_state->getValue('launch_user');

    if (!empty($delete_resources)) {
      $resources = $cloud_server_template->get('field_launch_resources')->getValue();
      if (!empty($resources)) {
        $entities = [];
        $schedules = [];
        foreach ($resources ?: [] as $resource) {
          $type = $resource['item_key'];
          $id = $resource['item_value'];
          $entity = $this->entityTypeManager
            ->getStorage($type)
            ->load($id);
          if (empty($entity)) {
            continue;
          }

          $entities[] = $entity;

          // Collect schedule entities.
          if ($type === 'k8s_deployment' || $type === 'k8s_pod') {
            $schedule_entities = $this->entityTypeManager
              ->getStorage('k8s_schedule')
              ->loadByProperties([
                'cloud_context' => $cloud_server_template->getCloudContext(),
                'launch_template_name' => $cloud_server_template->getName(),
                'namespace_name' => $entity->getNamespace(),
                'resource_name' => $entity->getName(),
              ]);

            $schedules = array_merge($schedules, array_values($schedule_entities));
          }
        }
        if (!empty($entities)) {
          $this->k8sService->deleteResourcesWithEntities($entities, TRUE);
        }

        // Remove schedules.
        $this->k8sService->deleteSchedules($schedules);

        $cloud_server_template->get('field_launch_resources')->setValue(NULL);
      }
    }

    $route = [
      'route_name' => 'entity.cloud_server_template.canonical',
      'params' => [
        'cloud_server_template' => $cloud_server_template->id(),
        'cloud_context' => $cloud_server_template->getCloudContext(),
      ],
    ];

    $this->k8sService->setCloudContext($cloud_server_template->getCloudContext());

    if ($source_type === 'git') {
      $this->launchByGitYamlFiles($cloud_server_template, $form_state, $launch_uid);
      return $route;
    }
    $object_types = $this->k8sService->supportedCloudServerTemplates();
    if (empty($object_types[$k8s_obj])) {
      $this->messenger->addError($this->t('@object launch not supported.', ['@object' => $k8s_obj]));
      return $route;
    }

    $errors = $this->checkExistence(
      $cloud_server_template->get('field_detail')->value,
      $cloud_server_template->getCloudContext(),
      $cloud_server_template->get('field_namespace')->value
    );

    if (!empty($errors)) {
      $resources = '';
      foreach ($errors as $error) {
        $resources .= "<li>{$error->render()}</li>";
      }

      $message = $this->formatPlural(count($errors),
        'The following resource still exists. Please launch again after completely deleting it.<ul>@resources</ul>',
        'The following resources still exist. Please launch again after completely deleting them.<ul>@resources</ul>', [
          '@resources' => Markup::create($resources),
        ]);

      $this->messenger->addError($message);

      return $route;
    }

    return $this->launchK8sResources($cloud_server_template, $launch_uid);
  }

  /**
   * {@inheritdoc}
   */
  public function buildListHeader() {
    return [
      [
        'data' => $this->t('Namespace'),
        'specifier' => 'field_namespace',
        'field' => 'field_namespace',
      ], [
        'data' => $this->t('Object'),
        'specifier' => 'field_object',
        'field' => 'field_object',
      ], [
        'data' => $this->t('Enable time scheduler'),
        'specifier' => 'field_enable_time_scheduler',
        'field' => 'field_enable_time_scheduler',
      ], [
        'data' => $this->t('Workflow status'),
        'specifier' => 'field_workflow_status',
        'field' => 'field_workflow_status',
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildListRow(CloudServerTemplateInterface $entity) {
    $row['field_namespace'] = [
      'data' => $this->entityLinkRenderer->renderViewElement(
        $entity->get('field_namespace')->value,
        'k8s_namespace',
        'name',
        []
      ),
    ];
    $row['field_object'] = [
      'data' => $this->renderField($entity, 'field_object'),
    ];
    $row['field_enable_time_scheduler'] = [
      'data' => $this->renderField($entity, 'field_enable_time_scheduler'),
    ];
    $row['field_workflow_status'] = [
      'data' => $this->renderField($entity, 'field_workflow_status'),
    ];
    return $row;
  }

  /**
   * Build the launch form for K8s.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state if launch is called from a form.
   */
  public function buildLaunchForm(array &$form, FormStateInterface $form_state): void {
    $form_object = $form_state->getFormObject();
    $entity = $form_object->getEntity();
    $source_type = $entity->get('field_source_type')->value;

    try {
      $form['#parents'] = [];

      $form_display = $this->entityTypeManager
        ->getStorage('entity_form_display')
        ->load('cloud_server_template.k8s.default');
      $widget_cloud_context = $form_display->getRenderer('cloud_context');
      $items_cloud_context = $entity->get('cloud_context');
      $items_cloud_context->filterEmptyItems();
      $form['cloud_context'] = $widget_cloud_context->form($items_cloud_context, $form, $form_state);
      $form['cloud_context']['widget'][0]['value']['#title'] = $this->t('K8s Cluster');
      $form['cloud_context']['widget'][0]['value']['#type'] = 'select';
      $option = $this->k8sService->clusterAllowedValues();
      $option['Automatic selection'] = 'Automatic selection';
      $form['cloud_context']['widget'][0]['value']['#options'] = $option;
      $form['cloud_context']['widget'][0]['value']['#ajax'] = [
        'callback' => 'k8s_ajax_callback_get_fields',
      ];
      unset(
        $form['cloud_context']['widget'][0]['value']['#size'],
        $form['cloud_context']['widget'][0]['value']['#description']
      );

      $widget_namespace = $form_display->getRenderer('field_namespace');
      $items_namespace = $entity->get('field_namespace');
      $items_namespace->filterEmptyItems();
      $form['field_namespace'] = $widget_namespace->form($items_namespace, $form, $form_state);

      $resources = $entity->get('field_launch_resources')->getValue();
      if (!empty($resources)) {
        k8s_create_resources_message($form, $resources);
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
    }

    $fieldsets_def = [
      [
        'name' => 'k8s',
        'title' => $this->t('K8s'),
        'open' => TRUE,
        'fields' => [
          'cloud_context',
          'field_namespace',
        ],
      ],
    ];

    // Check if entity has field,
    // Field might be missing if update has not been performed.
    if ($entity->hasField('field_launch_template_parameters')) {
      // Add editable KVP for launch_template_parameters,
      // Use the default renderer and display data in a fieldset.
      $widget_launch_template_parameters = $form_display->getRenderer('field_launch_template_parameters');
      $items_launch_template_parameters = $entity->get('field_launch_template_parameters');
      $items_launch_template_parameters->filterEmptyItems();
      $form['field_launch_template_parameters'] = $widget_launch_template_parameters->form($items_launch_template_parameters, $form, $form_state);
      $form['field_launch_template_parameters']['widget']['#title'] = NULL;

      // Add the fieldset definition.
      $fieldsets_def[] = [
        'name' => 'launch_template_parameters',
        'title' => t('Launch template parameters'),
        'open' => TRUE,
        'fields' => [
          'field_launch_template_parameters',
        ],
      ];
    }

    $git_available = $this->k8sService->isGitCommandAvailable();
    if (!$git_available && $source_type === 'git') {
      $this->messenger->addError(t('A git command could not be found on the server.  Please install the git command and try again.'));
      $this->handleError($form, $fieldsets_def);
      return;
    }
    if ($source_type === 'git') {

      $git_url_value = $entity->get('field_yaml_url')->getValue();
      $git_url = $git_url_value[0]['uri'];
      $git_branch_value = $entity->get('field_git_branch')->getValue();
      $git_branch = !empty($git_branch_value) ? $git_branch_value[0]['value'] : '';
      $git_paths = $entity->get('field_git_path')->getValue();
      $git_username = $entity->get('field_git_username')->value;
      $git_password = $entity->get('field_git_password')->value;

      $tmp_dir = $this->fileSystem->getTempDirectory();
      $name = Html::getId($entity->getName() . ' ' . $entity->id());
      $dir_name = $this->fileSystem->getDestinationFilename($tmp_dir . "/$name", FileSystemInterface::EXISTS_RENAME);

      // Make the temporary working directory.
      $result = $this->fileSystem->mkdir($dir_name);

      if (!$result) {
        $this->messenger->addError($this->t('Unable to create the working directory.'));
        $this->handleError($form, $fieldsets_def);
        return;
      }
      if ($git_username) {
        $account = rawurlencode($git_username);
        if ($git_password) {
          $account .= ":" . rawurlencode($git_password);
        }
        $git_url = str_replace("://", "://$account@", $git_url);
      }
      $output = [];
      $return_var = 0;
      if (!empty($git_branch)) {
        exec("git clone -b $git_branch $git_url $dir_name", $output, $return_var);
      }
      else {
        exec("git clone $git_url $dir_name", $output, $return_var);
      }
      if ($return_var !== 0) {
        $this->messenger->addError($this->t('Unable to clone the git repository.'));
        $this->handleError($form, $fieldsets_def);
        return;
      }
      $match = '/.*/';
      $files = [];
      try {
        foreach ($git_paths ?: [] as $path) {
          $files = array_merge($files, $this->fileSystem->scanDirectory($dir_name . $path['value'], $match));
        }
      }
      catch (NotRegularDirectoryException $e) {
        $this->messenger->addError($this->t('Git Resource Path might not be correct.'));
        $this->handleError($form, $fieldsets_def);
        return;
      }

      if (!empty($files)) {
        usort($files, static function ($file1, $file2) {
          $dir1 = dirname($file1->uri);
          $dir2 = dirname($file2->uri);

          if ($dir1 === $dir2) {
            return ($file1 > $file2) ? 1 : -1;
          }
          return ($dir1 > $dir2) ? 1 : -1;
        });

        $templates = $this->k8sService->supportedCloudServerTemplates();
        $extensions = $this->configFactory->get('k8s.settings')
          ->get('k8s_yaml_file_extensions');
        $regex = '/\.(' . preg_replace('/ +/', '|', preg_quote($extensions)) . ')$/i';

        foreach ($files ?: [] as $file) {
          $message = NULL;
          $validated = FALSE;
          $file_contents = file_get_contents($file->uri);
          if (!$file_contents) {
            $message = $this->t('Unable to get contents from the file %file.', ['%file' => $file->filename]);
            continue;
          }
          if (!preg_match($regex, $file->filename)) {
            continue;
          }

          try {
            $yamls = $this->k8sService->decodeMultipleDocYaml(
              // We can render the Yaml even if we
              // do not have launch_template_parameters.
              // This way, we bypass errors thrown by decoding
              // and the below validation will show a more helpful error.
              $this->renderYaml($file_contents, $this->getLaunchTemplateParameters($entity))
            );

            // Validate YAML files.
            foreach ($yamls ?: [] as $yaml) {
              if (!is_array($yaml)) {
                $message = $this->t("The file %file isn't YAML format.", ['%file' => $file->filename]);
                break;
              }

              // Check if there have been any placeholders,
              // replaced by NULL during rendering.
              $undefined_placeholders = $this->checkUndefinedPlaceholders($file_contents, $this->getLaunchTemplateParameters($entity));
              if (!empty($undefined_placeholders)) {
                $missing_parameters = '';
                foreach ($undefined_placeholders as $missing_parameter) {
                  $missing_parameters .= "<li>{$missing_parameter}</li>";
                }
                $message = $this->t('The file <strong><a href=":url">%file</a></strong> contains undefined parameters:<ul>@missing_parameters</ul>',
                [
                  ':url' => $entity->toUrl('edit-form', ['absolute' => TRUE])->toString(),
                  '%file' => $file->filename,
                  '@missing_parameters' => Markup::create($missing_parameters),
                ]);
                break;
              }

              if (!isset($yaml['kind'])) {
                $message = $this->t("No 'Kind' element found in the file %file.", ['%file' => $file->filename]);
                break;
              }

              $kind = $yaml['kind'];
              $object = array_search($kind, $templates);
              if ($object === FALSE) {
                $message = $this->t("Unsupported 'Kind' element in the file %file.", ['%file' => $file->filename]);
                break;
              }

              $validated = TRUE;
            }
          }
          catch (InvalidDataTypeException $e) {
            $message = $this->t('The file %file fails to decode: @exception', [
              '%file' => $file->filename,
              '@exception' => $e->getMessage(),
            ]);
          }

          $directory = str_replace($dir_name, '', dirname($file->uri));

          if (!isset($form['yaml_files'][$directory])) {
            $form['yaml_files'][$directory] = [
              '#type' => 'details',
              '#title' => $directory,
              '#open' => TRUE,
              'files' => [],
            ];
          }
          $form['yaml_files'][$directory]['files'][] = [
            '#type' => 'details',
            '#title' => $file->filename,
            '#open' => !$validated,
          ];
          $idx = count($form['yaml_files'][$directory]['files']) - 1;
          if (!$validated && !empty($message)) {
            $form['yaml_files'][$directory]['files'][$idx] += [
              'message' => [
                '#theme' => 'status_messages',
                '#message_list' => ['error' => [$message]],
              ],
              '#attributes' => [
                'class' => ['error', 'has-error'],
              ],
            ];
          }
          $form['yaml_files'][$directory]['files'][$idx]["launch_resource_$idx"] = [
            '#type' => 'checkbox',
            '#title' => $this->t('Launch this resource'),
            '#default_value' => $validated,
          ];
          if (!$validated && !empty($message)) {
            $form['yaml_files'][$directory]['files'][$idx]["launch_resource_$idx"]['#attributes'] = ['class' => ['error']];
          }
          $form['yaml_files'][$directory]['files'][$idx] += [
            "yaml_file_name_$idx" => [
              '#type' => 'hidden',
              '#default_value' => $file->filename,
            ],
            "yaml_file_content_$idx" => [
              '#type' => 'textarea',
              '#default_value' => $file_contents,
              '#attributes' => ['readonly' => 'readonly'],
            ],
          ];
        }
      }

      // If we do not have any launch resources, we display a warning.
      if (!isset($form['yaml_files'])) {
        $this->messenger->addWarning($this->t('There were no launch resources found. Please check <a href=":url">%label</a> for the Git repository or YAML URL settings and try again.', [
          ':url' => $entity->toUrl('edit-form', ['absolute' => TRUE])->toString(),
          '%label' => $entity->label(),
        ]));
        $this->handleError($form, $fieldsets_def);
        return;
      }

      $fieldsets_def[] = [
        'name' => 'launch_resources',
        'title' => $this->t('Launch Resources'),
        'open' => TRUE,
        'fields' => [
          'yaml_files',
        ],
      ];

      $this->fileSystem->deleteRecursive($dir_name);
    }

    $resources = $entity->get('field_launch_resources')->getValue();
    if (!empty($resources)) {
      $form['field_delete_all_resources'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Delete following resources'),
        '#default_value' => FALSE,
      ];

      $fieldsets_def[] = [
        'name' => 'delete_resources',
        'title' => $this->t('Delete Resources'),
        'open' => TRUE,
        'fields' => [
          'field_delete_all_resources',
          'confirm_message',
        ],
      ];
    }

    // Show only if Launch form.
    if (!empty($entity->hasField('field_workflow_status')
      && !empty($entity->field_workflow_status))
      && $entity->get('field_workflow_status')->value === CloudServerTemplateInterface::APPROVED) {
      $user = User::load($this->currentUser->id());
      $form['launch_user'] = [
        '#type'          => 'entity_autocomplete',
        '#target_type'   => 'user',
        '#title'         => $this->t('Launch as'),
        '#size'          => 60,
        '#default_value' => $user,
        '#weight'        => 0,
        '#required'      => TRUE,
        '#selection_handler' => 'default:launch_template_user',
        '#element_validate' => [
          [EntityAutocomplete::class, 'validateEntityAutocomplete'],
          [LaunchTemplateUserSelection::class, 'validateUser'],
        ],
      ];

      $fieldsets_def[] = [
        'name' => 'username',
        'title' => t('Username'),
        'open' => TRUE,
        'fields' => [
          'launch_user',
        ],
      ];
    }

    $this->cloudService->reorderForm($form, $fieldsets_def);
    $form['description']['#weight'] = (!empty($form['username']) ? $form['username']['#weight'] : count($fieldsets_def)) + 1;

  }

  /**
   * Render a launch template entity field.
   *
   * @param \Drupal\cloud\Entity\CloudServerTemplateInterface $entity
   *   The launch template entity.
   * @param string $field_name
   *   The field to render.
   * @param string $view
   *   The view to render.
   *
   * @return array
   *   A fully loaded render array for that field or empty array.
   */
  private function renderField(CloudServerTemplateInterface $entity, $field_name, $view = 'default') {
    $field = [];
    if ($entity->hasField($field_name)) {
      $field = $entity->get($field_name)->view($view);
      // Hide the label.
      $field['#label_display'] = 'hidden';
    }
    return $field;
  }

  /**
   * Launch a K8s Resource from a launch template.
   *
   * @param \Drupal\cloud\Entity\CloudServerTemplateInterface $cloud_server_template
   *   Cloud launch template interface.
   * @param int $launch_uid
   *   The launch user ID.
   *
   * @return array
   *   The route to redirect after launch.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   *   Thrown when unable to launch deployment.
   */
  private function launchK8sResources(CloudServerTemplateInterface $cloud_server_template, $launch_uid): array {
    $object_types = $this->k8sService->supportedCloudServerTemplates();

    // If we have twig values within this template, we replace.
    $rendered_file_contents
      = !empty($cloud_server_template->get('field_detail'))
      && !empty($cloud_server_template->get('field_detail')->value)
      ? $this->renderYaml(
        $cloud_server_template->get('field_detail')->value,
        $this->getLaunchTemplateParameters($cloud_server_template))
      : '';

    $yamls = $this->k8sService->decodeMultipleDocYaml($rendered_file_contents);

    foreach ($yamls ?: [] as $yaml) {
      // Add owner uid to annotations.
      $yaml['metadata']['annotations'][K8sEntityBase::ANNOTATION_CREATED_BY_UID] = $launch_uid;
      if ($yaml['kind'] === 'Deployment') {
        // Add owner uid to pod template.
        $yaml['spec']['template']['metadata']['annotations'][K8sEntityBase::ANNOTATION_CREATED_BY_UID] = $launch_uid;
      }

      $yaml['metadata']['annotations'][K8sEntityBase::ANNOTATION_LAUNCHED_APPLICATION_ID] = $cloud_server_template->id();
      $kind = $yaml['kind'];
      $object_type = array_search($kind, $object_types);
      $entity_type_id = "k8s_$object_type";

      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      $id_plural = $entity_type->get('id_plural');
      // We can't use empty() here since namespaceable can be
      // one of two values such as FALSE or NULL.
      $namespaceable = $entity_type->get('namespaceable') === NULL;

      // Get the name of method create<ResourceName>.
      $short_label = self::getCamelCaseWithWhitespace($object_type);
      $name_camel = self::getCamelCaseWithoutWhitespace($short_label);
      $create_method_name = "create${name_camel}";

      // Get the name of method update<ResourceName>s.
      $name_plural_camel = !empty($id_plural) ? self::getCamelCaseWithoutWhitespace($entity_type->getPluralLabel()) : '';
      $update_method_name = "update${name_plural_camel}";

      try {
        if ($namespaceable) {
          $namespace = $cloud_server_template->get('field_namespace')->value;
          $result = $this->k8sService->$create_method_name(
            $cloud_server_template->get('field_namespace')->value,
            $yaml
          );

          if ($kind === 'Pod' || $kind === 'Deployment') {
            $enable_time_scheduler = $cloud_server_template->get('field_enable_time_scheduler')->value;
            if (!empty($enable_time_scheduler)) {
              $timestamp = time();
              $resource_name = $result['metadata']['name'];
              $start_hour = $cloud_server_template->get('field_startup_time_hour')->value;
              $start_minute = $cloud_server_template->get('field_startup_time_minute')->value;
              $stop_hour = $cloud_server_template->get('field_stop_time_hour')->value;
              $stop_minute = $cloud_server_template->get('field_stop_time_minute')->value;
              $schedule = K8sSchedule::create([
                'cloud_context' => $cloud_server_template->getCloudContext(),
                'name' => $namespace . '_' . $resource_name,
                'kind' => $kind,
                'namespace_name' => $namespace,
                'resource_name' => $resource_name,
                'launch_template_name' => $cloud_server_template->getName(),
                'start_hour' => $start_hour,
                'start_minute' => $start_minute,
                'start_time' => $this->getTimeFormat($start_hour, $start_minute),
                'stop_hour' => $stop_hour,
                'stop_minute' => $stop_minute,
                'stop_time' => $this->getTimeFormat($stop_hour, $stop_minute),
                'manifest' => Yaml::encode($yaml),
                'created' => $timestamp,
                'changed' => $timestamp,
                'refreshed' => $timestamp,
              ]);
              $schedule->save();
            }
          }
        }
        else {
          $result = $this->k8sService->$create_method_name($yaml);
        }

        $this->k8sService->$update_method_name([
          'metadata.name' => $result['metadata']['name'],
        ], FALSE);

        $params = [
          'cloud_context' => $cloud_server_template->getCloudContext(),
          'name' => $result['metadata']['name'],
        ];

        $params += !empty($namespaceable) ? ['namespace' => $result['metadata']['namespace']] : [];

        // Update creation_yaml field of entity.
        $entities = $this->entityTypeManager->getStorage($entity_type_id)->loadByProperties($params);
        if (!empty($entities)) {
          $entity = reset($entities);

          // Update creation_yaml.
          if (method_exists($entity, 'setCreationYaml')) {
            unset($yaml['metadata']['annotations'][K8sEntityBase::ANNOTATION_LAUNCHED_APPLICATION_ID]);
            $entity->setCreationYaml(Yaml::encode($yaml));
            $entity->save();
          }

          // Update owner uid.
          $entity->setOwnerId($launch_uid);
          $entity->save();

          $this->messenger->addStatus($this->t('The @type <a href=":url">%label</a> has been launched.', [
            '@type' => $entity_type->getSingularLabel(),
            ':url' => $entity->toUrl('canonical')->toString(),
            '%label' => $result['metadata']['name'],
          ]));

          $this->logger('k8s')->info($this->t('@type: created %label.', [
            '@type' => $entity_type->getSingularLabel(),
            '%label' => $result['metadata']['name'],
          ]), [
            'link' => $entity->toLink($this->t('View'))->toString(),
          ]);

          $cloud_server_template->get('field_launch_resources')->appendItem([
            'item_key' => $entity->getEntityTypeId(),
            'item_value' => $entity->id(),
          ]);

          // Identify whether called by 'launchByGitYamlFiles'.
          // Parameters are 3 when called by the function.
          // And the cloud launch template shouldn't be saved.
          if (count(func_get_args()) === 2) {
            $cloud_server_template->setValidationRequired(FALSE);
            $cloud_server_template->save();
          }
        }
        else {
          $this->messenger->addStatus($this->t('The @type %label has been launched.', [
            '@type' => $short_label,
            '%label' => $result['metadata']['name'],
          ]));

          $this->logger('k8s')->info($this->t('@type: launched %label.', [
            '@type' => $short_label,
            '%label' => $result['metadata']['name'],
          ]));
        }

      }
      catch (K8sServiceException
      | EntityStorageException
      | EntityMalformedException $e) {

        $this->processOperationErrorStatus($cloud_server_template, 'launched');

        $route = [
          'route_name' => 'entity.cloud_server_template.canonical',
          'params' => [
            'cloud_server_template' => $cloud_server_template->id(),
            'cloud_context' => $cloud_server_template->getCloudContext(),
          ],
        ];

        return $route;
      }
    }

    if (count($yamls) === 1) {
      $route = [
        'route_name' => "view.$entity_type_id.list",
        'params' => ['cloud_context' => $cloud_server_template->getCloudContext()],
      ];
    }
    else {
      $route = [
        'route_name' => 'entity.cloud_server_template.canonical',
        'params' => [
          'cloud_server_template' => $cloud_server_template->id(),
          'cloud_context' => $cloud_server_template->getCloudContext(),
        ],
      ];
    }

    return $route;
  }

  /**
   * Launch by YAML files in git repository.
   *
   * @param \Drupal\cloud\Entity\CloudServerTemplateInterface $cloud_server_template
   *   Cloud launch template interface.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state if launch is called from a form.
   * @param int $launch_uid
   *   The launch user ID.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\TypedData\Exception\ReadOnlyException
   */
  private function launchByGitYamlFiles(CloudServerTemplateInterface $cloud_server_template, FormStateInterface $form_state = NULL, $launch_uid): void {
    $values = $form_state->getValues();

    $launch_resources = [];
    $yaml_file_contents = [];
    $yaml_file_names = [];
    $pattern = '/\d+/';
    foreach ($values as $key => $value) {
      if (strpos($key, 'launch_resource') !== FALSE) {
        $match = [];
        preg_match($pattern, $key, $match);
        if (!empty($match)) {
          $idx = $match[0];
          $launch_resources[$idx] = $value;
        }
      }
      if (strpos($key, 'yaml_file_content') !== FALSE) {
        $match = [];
        preg_match($pattern, $key, $match);
        if (!empty($match)) {
          $idx = $match[0];
          $yaml_file_contents[$idx] = $value;
        }
      }
      if (strpos($key, 'yaml_file_name') !== FALSE) {
        $match = [];
        preg_match($pattern, $key, $match);
        if (!empty($match)) {
          $idx = $match[0];
          $yaml_file_names[$idx] = $value;
        }
      }
    }

    $objects = [];
    if (!empty($yaml_file_contents) && !empty($launch_resources) && !empty($yaml_file_names)) {
      $templates = $this->k8sService->supportedCloudServerTemplates();
      $errors = [];
      foreach ($yaml_file_contents as $idx => $file_contents) {
        if (empty($launch_resources[$idx])) {
          continue;
        }

        try {
          // Render YAML.
          // Even if there is nothing to render,
          // this will not affect the functionality.
          $rendered_yaml_contents = $this->renderYaml($file_contents, $this->getLaunchTemplateParameters($cloud_server_template));
          $yamls = $this->k8sService->decodeMultipleDocYaml($rendered_yaml_contents);
          $validated = FALSE;
          $object = NULL;

          // Validate yamls.
          foreach ($yamls ?: [] as $yaml) {
            if (!is_array($yaml)) {
              $this->messenger->addError($this->t("The file %file isn't YAML format.", ['%file' => $yaml_file_names[$idx]]));
              break;
            }

            // Check if there have been any placeholders
            // replaced by NULL during rendering.
            $undefined_placeholders = $this->checkUndefinedPlaceholders($file_contents, $this->getLaunchTemplateParameters($cloud_server_template));
            if (!empty($undefined_placeholders)) {
              $missing_parameters = '';
              foreach ($undefined_placeholders as $missing_parameter) {
                $missing_parameters .= "<li>{$missing_parameter}</li>";
              }
              $this->messenger->addError($this->t('The file <strong><a href=":url">%file</a></strong> contains undefined parameters:<ul>@missing_parameters</ul>',
              [
                ':url' => $cloud_server_template->toUrl('edit-form', ['absolute' => TRUE])->toString(),
                '%file' => $yaml_file_names[$idx],
                '@missing_parameters' => Markup::create($missing_parameters),
              ]));
              break;
            }

            if (!isset($yaml['kind'])) {
              $this->messenger->addError($this->t("No 'Kind' element found in the file %file.", ['%file' => $yaml_file_names[$idx]]));
              break;
            }

            $kind = $yaml['kind'];
            $object = array_search($kind, $templates);
            if ($object === FALSE) {
              $this->messenger->addError($this->t("Unsupported 'Kind' element in the file %file.", ['%file' => $yaml_file_names[$idx]]));
              break;
            }

            $validated = TRUE;
          }

          // Whenever we need to check something,
          // we should use the rendered $yaml string.
          $result = !empty($validated)
            ? $this->checkExistence(
              $rendered_yaml_contents,
              $cloud_server_template->getCloudContext(),
              $cloud_server_template->get('field_namespace')->value
            ) : [];

          if ($validated && !empty($result)) {
            $errors = array_merge($errors, $result);
            continue;
          }
          if ($validated) {
            $cloud_server_template->get('field_detail')->setValue($rendered_yaml_contents);
            $errors_before = count($this->messenger->messagesByType($this->messenger::TYPE_ERROR));
            // Put an extra 2nd param to identify called by this function.
            $this->launchK8sResources($cloud_server_template, $launch_uid, TRUE);
            $errors_after = count($this->messenger->messagesByType($this->messenger::TYPE_ERROR));
            if ($errors_before === $errors_after) {
              if (count($yamls) === 1) {
                $objects[] = $object;
              }
              else {
                $objects[] = 'mixed';
              }
            }
          }
        }
        catch (\Exception $e) {
          $this->messenger->addError($this->t('Invalid Yaml format in the file %file: %message', [
            '%file' => $yaml_file_names[$idx],
            '%message' => $e->getMessage(),
          ]));
        }
      }

      if (!empty($errors)) {
        $resources = '';
        foreach ($errors as $idx => $error) {
          $resources .= "<li>{$error->render()}</li>";
        }

        $message = $this->formatPlural(count($errors),
          'The following resource still exists. Please launch again after completely deleting it.<ul>@resources</ul>',
          'The following resources still exist. Please launch again after completely deleting them.<ul>@resources</ul>', [
            '@resources' => Markup::create($resources),
          ]);

        $this->messenger->addError($message);
      }

      $cloud_server_template->get('field_object')->setValue($objects);
      $cloud_server_template->get('field_detail')->setValue(NULL);
      $cloud_server_template->setValidationRequired(FALSE);
      $cloud_server_template->save();
    }
  }

  /**
   * Handling error on launch form.
   *
   * @param array $form
   *   Form elements.
   * @param array $fieldsets_def
   *   Field set definitions.
   */
  private function handleError(array &$form, array $fieldsets_def): void {
    // Hide submit button and description message.
    $form['actions']['submit']['#access'] = FALSE;
    $form['description']['#access'] = FALSE;

    if (!empty($form['confirm_message'])) {
      unset($form['confirm_message']);
    }

    $this->cloudService->reorderForm($form, $fieldsets_def);
  }

  /**
   * Replaces {{ values }} with KVP values.
   *
   * @param string $yaml_file_contents
   *   Result of file_get_contents on a yaml file.
   * @param array $vars
   *   The array of KVP's to replace in the file.
   *
   * @return string
   *   The rendered yaml string to be decoded.
   */
  private function renderYaml(string $yaml_file_contents, array $vars): string {
    return $this->twig->renderInline($yaml_file_contents, $vars)->__toString();
  }

  /**
   * Returns launch_template_parameters.
   *
   * @param \CloudServerTemplateInterface $cloud_server_template
   *   The server template to retrieve launch parameters from.
   *
   * @return array
   *   The set of KVP's.
   */
  private function getLaunchTemplateParameters(CloudServerTemplateInterface $cloud_server_template): array {
    if (!$cloud_server_template->hasField('field_launch_template_parameters')) {
      return [];
    }
    $result = [];
    $parameters = $cloud_server_template->get('field_launch_template_parameters')->getValue();
    foreach ($parameters ?: [] as $parameter) {
      if (!empty($parameter['item_key']) && !empty($parameter['item_value'])) {
        $result[$parameter['item_key']] = $parameter['item_value'];
      }
    }
    return $result;
  }

  /**
   * Check the existence of resources.
   *
   * @param string $yaml_string
   *   The rendered YAML string.
   * @param string $cloud_context
   *   The cloud context.
   * @param string $namespace
   *   The namespace.
   *
   * @return array
   *   YAMLS with resources still existing.
   */
  private function checkExistence($yaml_string, $cloud_context, $namespace): array {

    $self = $this;

    $promise = new Promise(function () use (&$promise, $self, $yaml_string, $namespace) {
      $object_types = $self->k8sService->supportedCloudServerTemplates();
      $yamls = $self->k8sService->decodeMultipleDocYaml($yaml_string);

      for ($count = 0; $count < $self::CHECK_RETRY_COUNT; $count++) {
        if (empty($yamls)) {
          break;
        }

        foreach ($yamls ?: [] as $idx => $yaml) {
          $name = $yaml['metadata']['name'];
          $kind = $yaml['kind'];
          $object_type = array_search($kind, $object_types);
          $entity_type_id = "k8s_$object_type";
          $entity_type = $self->entityTypeManager->getDefinition($entity_type_id);
          // We can't use empty() here since namespaceable can be
          // one of two values such as FALSE or NULL.
          $namespaceable = $entity_type->get('namespaceable') === NULL;

          // Check existence.
          $result = $self->k8sService->hasResource($entity_type_id,
            !empty($namespaceable)
              ? [
                'metadata.namespace' => $namespace,
                'metadata.name' => $name,
              ] : [
                'metadata.name' => $name,
              ]);

          if (empty($result)) {
            unset($yamls[$idx]);
            continue;
          }

          $yamls[$idx]['entity_type_id'] = $entity_type_id;
        }

        sleep($self::CHECK_SLEEP_SECOND);
      }

      $promise->resolve($yamls);
    });

    $result = $promise->wait();

    $messages = [];

    foreach ($result ?: [] as $yaml) {
      $name = $yaml['metadata']['name'];
      $entity_type_id = $yaml['entity_type_id'];
      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      // We can't use empty() here since namespaceable can be
      // one of two values such as FALSE or NULL.
      $namespaceable = $entity_type->get('namespaceable') === NULL;

      $name_camel = $self::getCamelCaseWithoutWhitespace($entity_type->getPluralLabel());
      $update_method_name = "update${name_camel}";

      // Refresh resources.
      $this->k8sService->$update_method_name([
        'metadata.name' => $name,
      ], FALSE);

      $params = [
        'cloud_context' => $cloud_context,
        'name' => $name,
      ];

      $params += !empty($namespaceable) ? ['namespace' => $namespace] : [];

      $entities = $this->entityTypeManager->getStorage($entity_type_id)->loadByProperties($params);

      if (empty($entities)) {
        $messages[] = $this->t('@type: %name', [
          '@type' => $entity_type->getSingularLabel(),
          '%name' => $name,
        ]);
        continue;
      }

      $entity = reset($entities);

      if (!empty($namespaceable) && !($entity instanceof K8sNamespace)) {
        $namespaces = $this->entityTypeManager->getStorage('k8s_namespace')
          ->loadByProperties([
            'name' => $entity->getNamespace(),
            'cloud_context' => $entity->getCloudContext(),
          ]);
        if (!empty($namespaces)) {
          $namespace_entity = reset($namespaces);
          $messages[] = $this->t('@type: <a href=":url">%name</a> (Namespace: <a href=":namespace_url">%namespace</a>)', [
            '@type' => $entity->getEntityType()->getSingularLabel(),
            '%name' => $entity->getName(),
            ':url' => $entity->toUrl('canonical')->toString(),
            '%namespace' => $namespace_entity->getName(),
            ':namespace_url' => $namespace_entity->toUrl('canonical')->toString(),
          ]);
          continue;
        }
      }
      $messages[] = $this->t('@type: <a href=":url">%name</a>', [
        '@type' => $entity->getEntityType()->getSingularLabel(),
        '%name' => $entity->getName(),
        ':url' => $entity->toUrl('canonical')->toString(),
      ]);
    }

    return $messages;
  }

  /**
   * Helper function to check for undefined placeholders.
   *
   * @param string $yaml
   *   Unrendered yaml representation as string.
   * @param array $placeholders
   *   Array of KVP's to check against,
   *    can be obtained from field_launch_template_parameters.
   *
   * @return array
   *   Empty array if all placeholders have values
   *    or array of undefined placeholders to display.
   */
  private function checkUndefinedPlaceholders(string $yaml, array $placeholders): array {
    // Extract all twig placeholders from the $yaml string.
    $twig_pattern = '/\{{[^}]*\}}/';
    preg_match_all($twig_pattern, $yaml, $matches);
    $leftovers = [];
    // If there are placeholders found
    // check to see if a KVP has been provided.
    foreach (array_unique($matches[0]) ?: [] as $match) {
      // Check to see if a kvp has been provided for it.
      preg_match('~{{(.*?)}}~', $match, $token);
      // If not, return placeholder.
      if (empty($placeholders[str_replace(' ', '', $token[1])])) {
        $leftovers[] = $token[1];
      }
    }

    return $leftovers;
  }

}
