<?php

namespace Drupal\Tests\k8s\Functional\cloud_budget;

use Drupal\k8s\Controller\K8sCostsControllerBase;
use Drupal\Tests\k8s\Functional\K8sTestBase;
use Drupal\Component\Serialization\Yaml;

/**
 * Test K8s cloud cost calculator(K8sCloudCostStorageTest).
 *
 * @group Cloud
 */
class K8sCloudCostStorageTest extends K8sTestBase {

  public const K8S_CLOUD_COST_CALCULATOR_REPEAT_COUNT = 1;

  /**
   * The project.
   *
   * @var array
   */
  protected $project;

  /**
   * Set up test.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function setUp(): void {
    self::$modules[] = 'cloud_budget';
    parent::setUp();
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function getPermissions():array {
    $this->project = $this->createProjectTestFormData(1);
    $namespaces['name'] = $this->project[0]['name[0][value]'];
    $this->createNamespaceTestEntity($namespaces);
    $this->namespace = $namespaces['name'];

    return [
      'list k8s pod',
      'add k8s pod',
      'view any k8s pod',
      'edit any k8s pod',
      'delete any k8s pod',
      'list k8s namespace',
      'view any k8s namespace',
      'edit any k8s namespace',
      'add k8s namespace',
      'delete any k8s namespace',
      'launch cloud project',
      'add cloud projects',
      'list cloud project',
      'delete any cloud projects',
      'edit any cloud projects',
      'view any published cloud projects',
      'access cloud project revisions',
      'revert all cloud project revisions',
      'delete all cloud project revisions',
      'add cloud cost storage',
      'list cloud cost storage',
      'view cloud cost storage',
      'edit cloud cost storage',
      'delete cloud cost storage',
    ];
  }

  /**
   * K8s update resource storage test.
   *
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Behat\Mink\Exception\ExpectationException
   */
  public function testK8sUpdateResourceStorageTest(): void {
    // Set parameters.
    $cloud_context = $this->cloudContext;
    $k8s_service = \Drupal::service('k8s');
    $plugin = \Drupal::service('plugin.manager.cloud_cost_storage');

    // Create entity.
    $nodes = $this->createNodeTestFormData(self::K8S_CLOUD_COST_CALCULATOR_REPEAT_COUNT);
    $pods = $this->createPodsRandomTestFormData($this->namespace);
    foreach ($pods ?: [] as $pod) {
      $this->createPodTestEntity($pod);
    }
    foreach ($nodes ?: [] as $node) {
      $this->createNodeTestEntity($node);
    }

    // Add mock data.
    $this->getNamespaceResourceUsageMockData($pods);
    $this->getNodeResourceUsageMockData($nodes);
    $this->calculateCostPerNamespaceMockData();

    // Get result.
    $result_node = $k8s_service->getNodeResourceUsage($cloud_context);
    $result_pod = $k8s_service->getNamespaceResourceUsage($cloud_context, $this->namespace);
    $result = $k8s_service->calculateCostPerNamespace($cloud_context, $this->namespace);

    $this->assertEquals($result['resources']['cpu_usage'], $result_pod['cpu_usage']);
    $this->assertEquals($result['resources']['memory_usage'], $result_pod['memory_usage']);
    $this->assertEquals($result['resources']['cpu_capacity'], $result_node['cpu_capacity']);
    $this->assertEquals($result['resources']['memory_capacity'], $result_node['memory_capacity']);

    // Process item in queue.
    $queue_name = 'k8s_update_resources_queue';
    /** @var \Drupal\Core\Queue\QueueWorkerInterface $queue_worker */
    $queue_worker = \Drupal::service('plugin.manager.queue_worker')->createInstance($queue_name);
    $queue = \Drupal::queue($queue_name);
    // Clear the queue just in case.
    $queue->deleteQueue();

    // Update resource storage.
    $controller = \Drupal::classResolver()->getInstanceFromDefinition(K8sCostsControllerBase::class);
    $cost_count = count($controller->getEc2CostTypes());

    foreach ($plugin->getDefinitions() ?: [] as $id => $definition) {
      if ($id === 'K8s_cloud_cost_storage') {
        $plugin_id = $plugin->createInstance($id);
        $plugin_id->updateResourceStorageEntity();
      }
    }
    $this->assertSame($cost_count, $queue->numberOfItems(), 'Item was not added to the queue.');

    // Loop through the items and process them.
    $i = 0;
    while ($i < $cost_count) {
      $item = $queue->claimItem();
      $this->assertSame($item->data['k8s_method_name'], 'updateResourceStorageEntity',
        'Queue item that was imported function does not belong to the correct function.');
      $queue_worker->processItem($item->data);
      $queue->deleteItem($item);
      $i++;
    }

    $this->assertSame(0, $queue->numberOfItems(), 'Item was not removed from the queue.');

    // Get data from test entity.
    $resource_entities = \Drupal::entityTypeManager()
      ->getStorage('cloud_resource_storage')
      ->loadByProperties([]);

    $cost = 0;
    $resources = 0;
    foreach ($resource_entities ?: [] as $entity) {
      $cost = $entity->get('cost')->value;
      $resources = Yaml::decode($entity->get('resources')->value);
    }

    $this->assertEquals(round($cost, 3), round($result['cost'], 3));
    $this->assertEquals(round($resources['cpu_usage'], 3), round($result_pod['cpu_usage'], 3));
    $this->assertEquals(round($resources['memory_usage'], 3), round($result_pod['memory_usage'], 3));
  }

  /**
   * K8s cost resource storage test.
   *
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Behat\Mink\Exception\ExpectationException
   */
  public function testK8sCostResourceStorageTest(): void {
    // Set parameters.
    $cloud_context = $this->cloudContext;
    $k8s_service = \Drupal::service('k8s');
    $plugin = \Drupal::service('plugin.manager.cloud_cost_storage');

    // Create entity.
    $nodes = $this->createNodeTestFormData(self::K8S_CLOUD_COST_CALCULATOR_REPEAT_COUNT);
    $pods = $this->createPodsRandomTestFormData($this->namespace);
    foreach ($pods ?: [] as $pod) {
      $this->createPodTestEntity($pod);
    }
    foreach ($nodes ?: [] as $node) {
      $this->createNodeTestEntity($node);
    }

    // Add mock data.
    $this->getNamespaceResourceUsageMockData($pods);
    $this->getNodeResourceUsageMockData($nodes);
    $this->calculateCostPerNamespaceMockData();

    // Scenario No.1:
    // No project entity will expect no update to cost storage entity.
    // Update cost storage.
    foreach ($plugin->getDefinitions() ?: [] as $id => $definition) {
      if ($id === 'K8s_cloud_cost_storage') {
        $plugin_id = $plugin->createInstance($id);
        $plugin_id->updateCostStorageEntity();
      }
    }

    // Get result.
    $cost_entities = \Drupal::entityTypeManager()
      ->getStorage('cloud_cost_storage')
      ->loadByProperties([]);

    // Assertion.
    $this->assertEquals(count($cost_entities), 0);

    // Scenario No.2: If project entity exists and no resource entity,
    // it will expect updating resource storage entity.
    // Prepare project entity.
    $this->project['name'] = $this->namespace;
    $this->project['cloud_context'] = $cloud_context;
    $this->project['field_k8s_clusters'][0]['value'] = $cloud_context;
    $this->createProjectTestEntity($this->project);

    // Update cost storage.
    foreach ($plugin->getDefinitions() ?: [] as $id => $definition) {
      if ($id === 'K8s_cloud_cost_storage') {
        $plugin_id = $plugin->createInstance($id);
        $plugin_id->updateCostStorageEntity();
      }
    }

    // Get result.
    $resource_entities = \Drupal::entityTypeManager()
      ->getStorage('cloud_resource_storage')
      ->loadByProperties([]);

    // Assertion.
    $this->assertEquals(count($resource_entities), 6);

    // Scenario No.3:
    // If latest cost storage entity updated within 1 hour exists,
    // it will expect skipping the update of cost storage.
    // Prepare cost storage entity with nearby current time.
    $cost_storage = [
      'cloud_context' => $cloud_context,
      'payer' => $this->namespace,
      'cost' => 1.0,
      'resources' => 0,
      'refreshed' => time() - random_int(1, 60 * 59),
    ];
    $this->createCloudCostStorageTestEntity($cost_storage);

    // Update cost storage.
    foreach ($plugin->getDefinitions() ?: [] as $id => $definition) {
      if ($id === 'K8s_cloud_cost_storage') {
        $plugin_id = $plugin->createInstance($id);
        $plugin_id->updateCostStorageEntity();
      }
    }

    // Get result.
    $cost_entities = \Drupal::entityTypeManager()
      ->getStorage('cloud_cost_storage')
      ->loadByProperties([]);

    // Assertion.
    $this->assertEquals(count($cost_entities), 1);

    // Scenario No.4:
    // If latest cost storage entity updated within 1 hour does not exist,
    // it will expect updating cost storage referring resource storage.
    // Clean cost entity used in Scenario No.3.
    foreach ($cost_entities ?: [] as $cost_entity) {
      $cost_entity->delete();
    }

    // Prepare for cost and resource storage updated over one hour before.
    $cost_storage = [
      'cloud_context' => $cloud_context,
      'payer' => $this->namespace,
      'cost' => 1.0,
      'resources' => 0,
      'refreshed' => time() - random_int(60 * 60 * 2, 60 * 60 * 5),
    ];
    $this->createCloudCostStorageTestEntity($cost_storage);
    $result = $k8s_service->calculateCostPerNamespace($cloud_context, $this->namespace);
    $resource_storage = [
      'cloud_context' => $cloud_context,
      'payer' => $this->namespace,
      'cost' => $result['cost'],
      'resources' => Yaml::encode($result['resources']),
      'refreshed' => time() - random_int(60 * 60 * 2, 60 * 60 * 5),
    ];
    $this->createCloudResourceStorageTestEntity($resource_storage);

    // Update cost storage.
    foreach ($plugin->getDefinitions() ?: [] as $id => $definition) {
      if ($id === 'K8s_cloud_cost_storage') {
        $plugin_id = $plugin->createInstance($id);
        $plugin_id->updateCostStorageEntity();
      }
    }

    // Get result.
    $cost_entities = \Drupal::entityTypeManager()
      ->getStorage('cloud_cost_storage')
      ->loadByProperties([]);

    // Assertion.
    $this->assertEquals(count($cost_entities), 2);
    $cost_entity = array_pop($cost_entities);
    $cost = $cost_entity->get('cost')->value;
    $this->assertEquals(round($cost, 3), round($result['cost'], 3));
  }

}
