<?php

namespace Drupal\openstack\Service\Rest;

use Drupal\openstack\Service\OpenStackBatchOperations;
use Drupal\openstack\Service\OpenStackServiceInterface;
use Drupal\aws_cloud\Entity\Ec2\AwsCloudEntity;
use Drupal\aws_cloud\Entity\Ec2\AwsCloudEntityInterface;
use Drupal\cloud\Entity\CloudConfigInterface;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceBase;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Queue\QueueFactory;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Interacts with OpenStack using Native API.
 */
class OpenStackService extends CloudServiceBase implements OpenStackServiceInterface {

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Cloud context string.
   *
   * @var string
   */
  protected $cloudContext;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Guzzle client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The lock interface.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Credentials array.
   *
   * @var array
   */
  protected $credentials;

  /**
   * Entity field manager interface.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Field type plugin manager.
   *
   * @var \Drupal\core\Field\FieldTypePluginManagerInterface
   */
  protected $fieldTypePluginManager;

  /**
   * The config factory.
   *
   * Subclasses should use the self::config() method, which may be overridden to
   * address specific needs when loading config, rather than this property
   * directly. See \Drupal\Core\Form\ConfigFormBase::config() for an example of
   * this.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Constructs a new OpenStackService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   An entity type manager instance.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Http\ClientFactory $client_factory
   *   The http client.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The lock interface.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   The queue factory.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\core\Field\FieldTypePluginManagerInterface $field_type_plugin_manager
   *   The field type plugin manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   A configuration factory.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager,
                              CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
                              ClientFactory $client_factory,
                              LockBackendInterface $lock,
                              QueueFactory $queue_factory,
                              EntityFieldManagerInterface $entity_field_manager,
                              FieldTypePluginManagerInterface $field_type_plugin_manager,
                              ConfigFactoryInterface $config_factory) {

    // The parent constructor takes care of $this->messenger object.
    parent::__construct();

    // Setup the entity type manager for querying entities.
    $this->entityTypeManager = $entity_type_manager;

    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->httpClient = $client_factory->fromOptions();
    $this->lock = $lock;
    $this->queueFactory = $queue_factory;
    $this->entityFieldManager = $entity_field_manager;
    $this->fieldTypePluginManager = $field_type_plugin_manager;
    $this->configFactory = $config_factory;
  }

  /**
   * Helper method to call API.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   Array interpretation of the response body.
   */
  public function getClient(array $credentials = []) {
    $http_method = $credentials['http_method'] ?? '';
    $endpoint = $credentials['endpoint'] ?? '';
    $data = $credentials['params'] ?? '';

    if (!empty($data)) {
      $params['body'] = json_encode($data);
    }

    $this->loadCredentials();
    $auth_token = $this->generateAuthToken();
    $params['headers']['X-Auth-Token'] = $auth_token;

    $params['headers']['Content-Type'] = !empty($credentials['Content-Type'])
      ? $credentials['Content-Type'] : 'application/json';

    return $this->request($http_method, $endpoint, $params);
  }

  /**
   * {@inheritdoc}
   */
  public function setCloudContext($cloud_context) {
    $this->cloudContext = $cloud_context;
    $this->cloudConfigPluginManager->setCloudContext($cloud_context);
  }

  /**
   * Load credentials.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   The credentials array.
   */
  private function loadCredentials(array $credentials = []) {
    if (!empty($credentials)) {
      $this->credentials = $credentials;
    }
    elseif (empty($this->credentials)) {
      $this->credentials = $this->cloudConfigPluginManager->loadCredentials();
    }

    return $this->credentials;
  }

  /**
   * Setup any default parameters for the Guzzle request.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   Array of parameters.
   */
  public function generateAuthToken(array $credentials = []) {
    $this->loadCredentials($credentials);

    $body = [
      'auth' => [
        'identity' => [
          'methods' => [0 => 'password'],
          'password' => [
            'user' => [
              'name' => $this->credentials['username'],
              'domain' => [
                'id' => $this->credentials['domain_id'],
              ],
              'password' => $this->credentials['password'],
            ],
          ],
        ],
        'scope' => [
          'project' => [
            'id' => $this->credentials['project_id'],
          ],
        ],
      ],
    ];

    $response = $this->request('POST', '/identity/v3/auth/tokens', ['json' => $body]);

    return $response['auth_token'] ?? $response;
  }

  /**
   * Get Instance List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance list API response.
   */
  public function describeInstances(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/compute/{$compute_version}/servers/detail",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    $results = $volume_id_arr = $security_groups = [];
    $instances = $response['body']['servers'] ?? [];

    foreach ($instances ?: [] as $key => $instance) {
      foreach ($instance['security_groups'] ?? [] as $key_group => $group) {
        $security_groups[$key_group] = [
          'GroupName' => $group['name'] ?? NULL,
        ];
      }

      // Get Volume Response.
      $volume_ids = $instance['os-extended-volumes:volumes_attached'];
      foreach ($volume_ids ?: [] as $volume_id_val) {
        $volume_id = $volume_id_val['id'];
        $volume_data = $this->describeVolumeById([], $volume_id);
        $image_id = $volume_data['volume_image_metadata']['image_id'] ?? NULL;
        $instance['image'] = empty($instance['image'])
          ? ['id' => $image_id] : [];
        $volume_id_arr = [
          'Ebs' => ['VolumeId' => $volume_id ?? ''],
        ];
      }

      // Get flavor Response.
      $flavor_id = $instance['flavor']['id'] ?? NULL;
      $flavor_data = $this->describeFlavorById([], $flavor_id);

      // Get port response.
      $port_result = $this->instancePortInterface([
        'InstanceId' => $instance['id'],
      ]);

      $port_ids = [];
      if (!empty($port_result)) {
        foreach ($port_result ?: [] as $port) {
          $port_ids[] = $port['port_id'];
        }
      }

      foreach ($instance['addresses'] ?: [] as $add_type => $address_val) {
        foreach ($address_val ?: [] as $add_value) {
          if ($add_type === 'private') {
            if ($add_value['OS-EXT-IPS:type'] === 'floating') {
              $public_ip = $add_value['addr'] ?? '';
            }
            elseif ($add_value['OS-EXT-IPS:type'] === 'fixed'
              && $add_value['version'] === OpenStackServiceInterface::IPV4) {
              $private_ip = $add_value['addr'] ?? '';
            }
          }
          elseif ($add_type === 'public'
            && $add_value['OS-EXT-IPS:type'] === 'fixed'
            && $add_value['version'] === OpenStackServiceInterface::IPV4) {
            $public_ip = $add_value['addr'] ?? '';
          }
        }
      }

      $results['Reservations'][$key] = [
        'OwnerId'    => $instance['tenant_id'],
        'Instances'  => [
          [
            'AmiLaunchIndex'      => NULL,
            'ImageId'             => $instance['image']['id'] ?? NULL,
            'InstanceId'          => $instance['id'] ?? NULL,
            'InstanceName'        => $instance['name'] ?? NULL,
            'InstanceType'        => $flavor_data['name'] ?? NULL,
            'KeyName'             => $instance['key_name'] ?? NULL,
            'LaunchTime'          => !empty($instance['OS-SRV-USG:launched_at'])
            ? str_replace('T', ' ', $instance['OS-SRV-USG:launched_at'])
            : str_replace('Z', '', str_replace('T', ' ', $instance['created'])),
            'Placement'           => ['AvailabilityZone' => $instance['OS-EXT-AZ:availability_zone'] ?? ''] ?? NULL,
            'PrivateDnsName'      => $instance['name'] ?? NULL,
            'PrivateIpAddress'    => $private_ip ?? '',
            'PublicIpAddress'     => $public_ip ?? '',
            'PublicDnsName'       => $instance['dns_name'] ?? NULL,
            'State'               => ['Name' => $instance['status'] === 'ACTIVE' ? 'running' : 'stopped'],
            'RootDeviceName'      => NULL,
            'RootDeviceType'      => NULL,
            'SecurityGroups'      => $security_groups ?? [],
            'BlockDeviceMappings' => [$volume_id_arr ?? []],
            'PortId'              => $port_ids ?? [],
          ],
        ],
      ];
    }

    return $results;
  }

  /**
   * Update Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance update API response.
   */
  public function modifyLaunchTemplate(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body['server'] = [
      "name" => $params['Name'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => "/compute/{$compute_version}/servers/{$params['InstanceId']}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $instance = $response['body']['server'] ?? [];

    $volume_id_arr = $security_groups = [];

    foreach ($instance['security_groups'] ?? [] as $key_group => $group) {
      $security_groups[$key_group] = [
        'GroupName' => $group['name'] ?? NULL,
      ];
    }

    // Get Volume Response.
    $volume_ids = $instance['os-extended-volumes:volumes_attached'] ?? [];
    foreach ($volume_ids ?: [] as $volume_id_val) {
      $volume_id = $volume_id_val['id'];
      $volume_data = $this->describeVolumeById([], $volume_id);
      $image_id = $volume_data['volume_image_metadata']['image_id'] ?? NULL;
      $instance['image'] = empty($instance['image'])
        ? ['id' => $image_id] : [];
      $volume_id_arr = [
        'Ebs' => ['VolumeId' => $volume_id ?? ''],
      ];
    }

    // Get flavor Response.
    $flavor_id = $instance['flavor']['id'] ?? NULL;
    $flavor_data = $this->describeFlavorById([], $flavor_id);

    // Get port response.
    $port_result = $this->instancePortInterface([
      'InstanceId' => $instance['id'],
    ]);

    $port_ids = [];
    if (!empty($port_result)) {
      foreach ($port_result ?: [] as $port) {
        $port_ids[] = $port['port_id'];
      }
    }

    foreach ($instance['addresses'] ?: [] as $address_val) {
      foreach ($address_val ?: [] as $add_value) {
        if ($add_value['OS-EXT-IPS:type'] === 'floating') {
          $public_ip = $add_value['addr'] ?? '';
        }
        if ($add_value['OS-EXT-IPS:type'] === 'fixed'
          && $add_value['version'] === OpenStackServiceInterface::IPV4) {
          $private_ip = $add_value['addr'] ?? '';
        }
      }
    }

    return [
      'Reservations' => [
        'OwnerId'    => $instance['tenant_id'],
        'Instances'  => [
          [
            'AmiLaunchIndex'      => NULL,
            'ImageId'             => $instance['image']['id'] ?? NULL,
            'InstanceId'          => $instance['id'] ?? NULL,
            'InstanceName'        => $instance['name'] ?? NULL,
            'InstanceType'        => $flavor_data['name'] ?? NULL,
            'KeyName'             => $instance['key_name'] ?? NULL,
            'LaunchTime'          => !empty($instance['OS-SRV-USG:launched_at'])
            ? str_replace('T', ' ', $instance['OS-SRV-USG:launched_at']) : NULL,
            'Placement'           => ['AvailabilityZone' => $instance['OS-EXT-AZ:availability_zone'] ?? ''] ?? NULL,
            'PrivateDnsName'      => $instance['name'] ?? NULL,
            'PrivateIpAddress'    => $private_ip ?? '',
            'PublicIpAddress'     => $public_ip ?? '',
            'PublicDnsName'       => $instance['dns_name'] ?? NULL,
            'State'               => ['Name' => $instance['status'] === 1 ? 'running' : 'stopped'],
            'RootDeviceName'      => NULL,
            'RootDeviceType'      => NULL,
            'SecurityGroups'      => $security_groups ?? [],
            'BlockDeviceMappings' => [$volume_id_arr ?? []],
            'PortId'              => $port_ids ?? [],
          ],
        ],
      ],
    ];
  }

  /**
   * Start Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   The Instance Start API response.
   */
  public function startInstances(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body = [
      'os-start' => NULL,
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/compute/{$compute_version}/servers/{$params['InstanceId']}/action",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Stop Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   The Instance Stop API response.
   */
  public function stopInstances(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body = [
      'os-stop' => NULL,
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/compute/{$compute_version}/servers/{$params['InstanceId']}/action",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Reboot Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   The Instance Reboot API response.
   */
  public function rebootInstances(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body['reboot'] = [
      'type' => 'HARD',
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/compute/{$compute_version}/servers/{$params['InstanceId']}/action",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Terminate|Delete Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   The Instance Terminate|Delete API response.
   */
  public function terminateInstance(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => "/compute/{$compute_version}/servers/{$params['InstanceId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * Run instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param array $tags
   *   The tags array.
   *
   * @return bool
   *   The create instance list API response.
   */
  public function runInstances(array $params = [], array $tags = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body = [
      'server' => [
        'name' => $params['Name'],
        'flavorRef' => $params['FlavorId'],
        'imageRef' => $params['ImageId'],
        'key_name' => $params['KeyName'],
        'availability_zone' => $params['Placement']['AvailabilityZone'],
        'security_groups' => $params['SecurityGroupIds'],
        'networks' => $params['NetworkInterfaces'],
      ],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/compute/{$compute_version}/servers",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Get Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image list API response.
   */
  public function describeImages(array $params = []) {
    $this->loadCredentials();
    $image_version = $this->credentials['image_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/image/{$image_version}/images",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);
    $images = $response['body']['images'] ?? [];

    $results = [];

    foreach ($images ?: [] as $key => $value) {
      $results['Images'][$key]['Architecture'] = NULL;
      $results['Images'][$key]['CreationDate'] = $value['created_at'];
      $results['Images'][$key]['ImageId'] = $value['id'];
      $results['Images'][$key]['ImageLocation'] = NULL;
      $results['Images'][$key]['ImageType'] = $value['disk_format'];
      $results['Images'][$key]['Public'] = $value['visibility'] === 'public' ? 1 : 0;
      $results['Images'][$key]['OwnerId'] = $value['owner'];
      $results['Images'][$key]['State'] = $value['status'] === 'active' ? 'available' : 'pending';
      $results['Images'][$key]['Name'] = $value['name'];
      $results['Images'][$key]['RootDeviceType'] = $value['container_format'];
    }

    return $results;
  }

  /**
   * Create Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image create API response.
   */
  public function createImage(array $params = []) {

    $this->loadCredentials();
    $image_version = $this->credentials['image_version'] ?? '';

    $body = [
      'container_format' => 'bare',
      'disk_format' => 'qcow2',
      'name' => $params['Name'],
      'visibility' => $params['Visibility'] ? 'public' : 'private',
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/image/{$image_version}/images",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return !empty($response) && !empty($response['body']) ? [
      'ImageId'          => $response['body']['id'],
      'name'             => $response['body']['name'],
      'disk_format'      => $response['body']['disk_format'],
      'container_format' => $response['body']['container_format'],
      'visibility'       => $response['body']['visibility'],
      'created_at'       => $response['body']['created_at'],
      'updated_at'       => $response['body']['updated_at'],
      'tags'             => $response['body']['tags'],
      'status'           => $response['body']['status'],
    ]
    : [];
  }

  /**
   * Update Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image update API response.
   */
  public function updateImage(array $params = []) {
    $this->loadCredentials();
    $image_version = $this->credentials['image_version'] ?? '';

    $body = [
      [
        'op' => 'replace',
        'path' => '/name',
        'value' => $params['Name'],
      ],
      [
        'op' => 'replace',
        'path' => '/visibility',
        'value' => $params['Visibility'],
      ],
    ];

    $credentials = [
      'http_method' => 'patch',
      'endpoint' => "/image/{$image_version}/images/{$params['ImageId']}",
      'params' => $body,
    ];

    $credentials['Content-Type'] = 'application/openstack-images-v2.1-json-patch';

    $response = $this->getClient($credentials);

    return !empty($response) && !empty($response['body']) ? [
      'ImageId'          => $response['body']['id'],
      'name'             => $response['body']['name'],
      'disk_format'      => $response['body']['disk_format'],
      'container_format' => $response['body']['container_format'],
      'visibility'       => $response['body']['visibility'],
      'created_at'       => $response['body']['created_at'],
      'updated_at'       => $response['body']['updated_at'],
      'tags'             => $response['body']['tags'],
      'status'           => $response['body']['status'],
    ]
    : [];
  }

  /**
   * Delete Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   The Image delete API response.
   */
  public function deregisterImage(array $params = []) {
    $this->loadCredentials();
    $image_version = $this->credentials['image_version'] ?? '';

    $credentials = [
      'http_method' => 'DELETE',
      'endpoint' => "/image/{$image_version}/images/{$params['ImageId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * Get key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The key pairs list API response.
   */
  public function describeKeyPairs(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $credentials = [
      'http_method' => 'GET',
      'endpoint' => "/compute/{$compute_version}/os-keypairs",
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $key_pairs = $response['body']['keypairs'] ?? [];

    $results['KeyPairs'] = [];
    foreach ($key_pairs ?: [] as $key => $key_pair) {
      $results['KeyPairs'][$key]['KeyName'] = $key_pair['keypair']['name'];
      $results['KeyPairs'][$key]['KeyFingerprint'] = $key_pair['keypair']['fingerprint'];
      $results['KeyPairs'][$key]['KeyMaterial'] = $key_pair['keypair']['public_key'];
    }

    return $results;
  }

  /**
   * Create key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The key pairs create API response.
   */
  public function createKeyPair(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body = [
      'keypair' => [
        'name' => $params['KeyName'],
      ],
    ];

    $credentials = [
      'http_method' => 'POST',
      'endpoint' => "/compute/{$compute_version}/os-keypairs",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return !empty($response)
      && !empty($response['body'])
      && !empty($response['body']['keypair'])
      ? [
        'KeyPairId' => NULL,
        'KeyName' => $response['body']['keypair']['name'],
        'KeyFingerprint' => $response['body']['keypair']['fingerprint'],
        'KeyMaterial' => $response['body']['keypair']['public_key'],
      ]
      : [];
  }

  /**
   * Import key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The key pairs import API response.
   */
  public function importKeyPair(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $body = [
      'keypair' => [
        'name' => $params['KeyName'],
        'public_key' => $params['PublicKeyMaterial'],
      ],
    ];

    $credentials = [
      'http_method' => 'POST',
      'endpoint' => "/compute/{$compute_version}/os-keypairs",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 200
      && !empty($response)
      && !empty($response['body'])
      && !empty($response['body']['keypair'])
      ? [
        'KeyPairId' => NULL,
        'KeyName' => $response['body']['keypair']['name'],
        'KeyFingerprint' => $response['body']['keypair']['fingerprint'],
        'KeyMaterial' => $response['body']['keypair']['public_key'],
      ]
      : [];
  }

  /**
   * Delete key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function deleteKeyPair(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $credentials = [
      'http_method' => 'DELETE',
      'endpoint' => "/compute/{$compute_version}/os-keypairs/{$params['KeyName']}",
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Get volume list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume list API response.
   */
  public function describeVolumes(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes/detail",
    ];
    $response = $this->getClient($credentials);

    $volumes = $response['body']['volumes'] ?? [];

    $results['Volumes'] = [];

    foreach ($volumes ?: [] as $key => $value) {
      $attachments = [];
      if (!empty($value['attachments'])) {
        foreach ($value['attachments'] as $attachment_key => $attachment) {
          $attachments[$attachment_key]['Device'] = $attachment['device'];
          $attachments[$attachment_key]['InstanceId'] = $attachment['server_id'];
          $attachments[$attachment_key]['VolumeId'] = $attachment['volume_id'];
        }
      }

      $results['Volumes'][$key]['Attachments'] = $attachments;
      $results['Volumes'][$key]['Name'] = $value['name'];
      $results['Volumes'][$key]['Description'] = $value['description'];
      $results['Volumes'][$key]['VolumeId'] = $value['id'];
      $results['Volumes'][$key]['State'] = $value['status'];
      $results['Volumes'][$key]['CreateTime'] = str_replace("T", " ", $value['created_at']);
      $results['Volumes'][$key]['SnapshotId'] = $value['snapshot_id'];
      $results['Volumes'][$key]['VolumeType'] = $value['volume_type'];
      $results['Volumes'][$key]['Size'] = $value['size'];
      $results['Volumes'][$key]['VirtualizationType'] = $value['replication_status'];
      $results['Volumes'][$key]['AvailabilityZone'] = $value['availability_zone'];
      $results['Volumes'][$key]['Encrypted'] = $value['encrypted'];
    }

    return $results;
  }

  /**
   * Create volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume create API response.
   */
  public function createVolume(array $params = []) {
    $this->loadCredentials();
    $volume_version = $this->credentials['volume_version'] ?? '';

    $body['volume'] = [
      'name' => $params['Name'],
      'size' => $params['Size'],
      'volume_type' => $params['VolumeType'],
      'availability_zone' => $params['AvailabilityZone'],
    ];

    if (!empty($params['SnapshotId'])) {
      $body['volume'] += ['snapshot_id' => $params['SnapshotId']];
    }

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $volume = $response['body']['volume'] ?? [];

    $attachments = [];

    if (!empty($volume['attachments'])) {
      foreach ($volume['attachments'] ?: [] as $attachment_key => $attachment) {
        $attachments[$attachment_key]['Device'] = $attachment['device'];
        $attachments[$attachment_key]['InstanceId'] = $attachment['server_id'];
        $attachments[$attachment_key]['VolumeId'] = $attachment['volume_id'];
      }
    }

    return [
      'Attachments'        => $attachments ?? [],
      'Name'               => $volume['name'] ?? NULL,
      'Description'        => $volume['description'] ?? NULL,
      'VolumeId'           => $volume['id'] ?? NULL,
      'State'              => $volume['status'] ?? NULL,
      'CreateTime'         => !empty($volume['created_at']) ? str_replace('T', ' ', $volume['created_at']) : NULL,
      'SnapshotId'         => $volume['snapshot_id'] ?? NULL,
      'VolumeType'         => $volume['volume_type'] ?? NULL,
      'Size'               => $volume['size'] ?? NULL,
      'VirtualizationType' => $volume['replication_status'] ?? NULL,
      'AvailabilityZone'   => $volume['availability_zone'] ?? NULL,
      'Encrypted'          => $volume['encrypted'] ?? NULL,
    ];
  }

  /**
   * Update volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume update API response.
   */
  public function updateVolume(array $params = []) {
    $this->loadCredentials();

    $body['volume'] = [
      'name' => $params['Name'],
    ];

    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'put',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes/{$params['VolumeId']}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $volume = $response['body']['volume'] ?? [];

    $attachments = [];

    if (!empty($volume['attachments'])) {
      foreach ($volume['attachments'] ?: [] as $attachment_key => $attachment) {
        $attachments[$attachment_key]['Device'] = $attachment['device'];
        $attachments[$attachment_key]['InstanceId'] = $attachment['server_id'];
        $attachments[$attachment_key]['VolumeId'] = $attachment['volume_id'];
      }
    }

    return [
      'Attachments'        => $attachments ?? [],
      'Name'               => $volume['name'] ?? NULL,
      'Description'        => $volume['description'] ?? NULL,
      'VolumeId'           => $volume['id'] ?? NULL,
      'State'              => $volume['status'] ?? NULL,
      'CreateTime'         => !empty($volume['created_at']) ? str_replace('T', ' ', $volume['created_at']) : NULL,
      'SnapshotId'         => $volume['snapshot_id'] ?? NULL,
      'VolumeType'         => $volume['volume_type'] ?? NULL,
      'Size'               => $volume['size'] ?? NULL,
      'VirtualizationType' => $volume['replication_status'] ?? NULL,
      'AvailabilityZone'   => $volume['availability_zone'] ?? NULL,
      'Encrypted'          => $volume['encrypted'] ?? NULL,
    ];
  }

  /**
   * Delete volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function deleteVolume(array $params = []) {
    $this->loadCredentials();

    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes/{$params['VolumeId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Attach volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function attachVolume(array $params = []) {
    $this->loadCredentials();
    $volume_version = $this->credentials['volume_version'] ?? '';

    $body['os-attach'] = [
      'instance_uuid' => $params['InstanceId'],
      'mountpoint' => $params['Device'],
    ];

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes/{$params['VolumeId']}/action",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Detach volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function detachVolume(array $params = []) {
    $this->loadCredentials();

    $body['os-detach'] = [
      'attachment_id' => $params['InstanceId'] ?? NULL,
    ];

    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes/{$params['VolumeId']}/action",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Get snapshot list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The snapshot list API response.
   */
  public function describeSnapshots(array $params = []) {

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/snapshots/detail",
    ];
    $response = $this->getClient($credentials);

    $snapshots = $response['body']['snapshots'] ?? [];

    $results['Snapshots'] = [];

    foreach ($snapshots ?: [] as $key => $snapshot) {
      $results['Snapshots'][$key]['Description'] = $snapshot['description'] ?? NULL;
      $results['Snapshots'][$key]['OwnerId'] = $snapshot['os-extended-snapshot-attributes:project_id'] ?? NULL;
      $results['Snapshots'][$key]['Progress'] = $snapshot['os-extended-snapshot-attributes:progress'] ?? NULL;
      $results['Snapshots'][$key]['name'] = $snapshot['name'] ?? NULL;
      $results['Snapshots'][$key]['SnapshotId'] = $snapshot['id'] ?? NULL;
      $results['Snapshots'][$key]['StartTime'] = $snapshot['created_at'] ?? NULL;
      $results['Snapshots'][$key]['State'] = $snapshot['status'] ?? NULL;
      $results['Snapshots'][$key]['VolumeId'] = $snapshot['volume_id'] ?? NULL;
      $results['Snapshots'][$key]['VolumeSize'] = $snapshot['size'] ?? NULL;
      $results['Snapshots'][$key]['Tags'] = $snapshot['metadata'] ?? [];
    }

    return $results;
  }

  /**
   * Create snapshot.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The snapshot create API response.
   */
  public function createSnapshot(array $params = []) {
    $this->loadCredentials();

    $body['snapshot'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
      'volume_id' => $params['VolumeId'],
    ];

    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/snapshots",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $snapshot = $response['body']['snapshot'] ?? [];

    return [
      'Description' => $snapshot['description'] ?? NULL,
      'OwnerId'     => $snapshot['os-extended-snapshot-attributes:project_id'] ?? NULL,
      'Progress'    => $snapshot['os-extended-snapshot-attributes:progress'] ?? NULL,
      'name'        => $snapshot['name'] ?? NULL,
      'SnapshotId'  => $snapshot['id'] ?? NULL,
      'StartTime'   => !empty($snapshot['created_at']) ? $snapshot['created_at'] : NULL,
      'State'       => $snapshot['status'] ?? NULL,
      'VolumeId'    => $snapshot['volume_id'] ?? NULL,
      'VolumeSize'  => $snapshot['size'] ?? NULL,
      'Tags'        => $snapshot['metadata'] ?? [],
    ];
  }

  /**
   * Update snapshot.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The snapshot update API response.
   */
  public function updateSnapshot(array $params = []) {
    $this->loadCredentials();

    $body['snapshot'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
    ];

    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'put',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/snapshots/{$params['SnapshotId']}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $snapshot = $response['body']['snapshot'] ?? [];

    return [
      'Description' => $snapshot['description'] ?? NULL,
      'OwnerId'     => $snapshot['os-extended-snapshot-attributes:project_id'] ?? NULL,
      'Progress'    => $snapshot['os-extended-snapshot-attributes:progress'] ?? NULL,
      'name'        => $snapshot['name'] ?? NULL,
      'SnapshotId'  => $snapshot['id'] ?? NULL,
      'StartTime'   => $snapshot['created_at'] ?? NULL,
      'State'       => $snapshot['status'] ?? NULL,
      'VolumeId'    => $snapshot['volume_id'] ?? NULL,
      'VolumeSize'  => $snapshot['size'] ?? NULL,
      'Tags'        => $snapshot['metadata'] ?? [],
    ];
  }

  /**
   * Delete snapshot.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function deleteSnapshot(array $params = []) {
    $this->loadCredentials();

    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/snapshots/{$params['SnapshotId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202;
  }

  /**
   * Get flavors.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Flavor list API response.
   */
  public function describeFlavors(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/compute/{$compute_version}/flavors",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['flavors'] ?? [];
  }

  /**
   * Get Flavor Detail of specific flavor Id.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param string $flavor_id
   *   The Flavor Id.
   *
   * @return array
   *   The Flavor list API response.
   */
  public function describeFlavorById(array $params = [], $flavor_id = '') {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/compute/{$compute_version}/flavors/{$flavor_id}",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['flavor'];
  }

  /**
   * Get volume Detail of specific volume Id.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param string $volume_id
   *   The Volume Id.
   *
   * @return array
   *   The volume list API response.
   */
  public function describeVolumeById(array $params = [], $volume_id = '') {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $volume_version = $this->credentials['volume_version'] ?? '';

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/volume/{$volume_version}/{$project_id}/volumes/{$volume_id}",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['volume'] ?? [];
  }

  /**
   * Get security group list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The security group list API response.
   */
  public function describeSecurityGroups(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'] ?? NULL;
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $endpoint = !empty($params['id'])
      ? ":{$api_endpoint_port}/{$network_version}/security-groups?project_id={$project_id}&id={$params['id']}"
      : ":{$api_endpoint_port}/{$network_version}/security-groups?project_id={$project_id}";

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $endpoint,
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $security_groups = $response['body']['security_groups'] ?? [];

    $results['SecurityGroups'] = [];

    foreach ($security_groups ?: [] as $key => $value) {

      $ip_permission_cnt = $ip_permission_egress_cnt = 0;

      $ip_permissions = $ip_permissions_egress = [];

      foreach ($value['security_group_rules'] ?: [] as $value_rules) {
        if (!empty($value_rules['direction']) && $value_rules['direction'] === 'egress') {

          $ip_permissions_egress[$ip_permission_egress_cnt] = [
            'RuleId'     => $value_rules['id'] ?? NULL,
            'FromPort'   => $value_rules['port_range_min'] ?? 1,
            'IpProtocol' => $value_rules['protocol'] ?? '-1',
            'ToPort'     => $value_rules['port_range_max'] ?? 65535,
          ];

          if (!empty($value_rules['remote_group_id'])) {
            $ip_permissions_egress[$ip_permission_egress_cnt] += [
              'UserIdGroupPairs' => [
                [
                  'GroupId' => $value_rules['remote_group_id'] ?? NULL,
                  'UserId'  => $value_rules['tenant_id'] ?? NULL,
                ],
              ],
            ];
          }

          if ((!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv6')
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions_egress[$ip_permission_egress_cnt] += [
              'Ipv6Ranges' => [
                [
                  'CidrIpv6'    => $value_rules['remote_ip_prefix'] ?? NULL,
                  'Description' => $value_rules['description'] ?? NULL,
                ],
              ],
            ];
          }

          if ((!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv4')
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions_egress[$ip_permission_egress_cnt] += [
              'IpRanges' => [
                [
                  'CidrIp'      => $value_rules['remote_ip_prefix'] ?? NULL,
                  'Description' => $value_rules['description'] ?? NULL,
                ],
              ],
            ];
          }

          $ip_permission_egress_cnt++;

        }
        else {
          $ip_permissions[$ip_permission_cnt] = [
            'RuleId'     => $value_rules['id'] ?? NULL,
            'FromPort'   => $value_rules['port_range_min'] ?? 1,
            'IpProtocol' => $value_rules['protocol'] ?? '-1',
            'ToPort'     => $value_rules['port_range_max'] ?? 65535,
          ];

          if (!empty($value_rules['remote_group_id'])) {
            $ip_permissions[$ip_permission_cnt] += [
              'UserIdGroupPairs' => [
                [
                  'GroupId' => $value_rules['remote_group_id'] ?? NULL,
                  'UserId'  => $value_rules['tenant_id'] ?? NULL,
                ],
              ],
            ];
          }

          if (!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv6'
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions[$ip_permission_cnt] += [
              'Ipv6Ranges' => [
                [
                  'CidrIpv6'    => $value_rules['remote_ip_prefix'] ?? NULL,
                  'Description' => $value_rules['description'] ?? NULL,
                ],
              ],
            ];
          }
          if (!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv4'
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions[$ip_permission_cnt] += [
              'IpRanges' => [
                [
                  'CidrIp'      => $value_rules['remote_ip_prefix'] ?? NULL,
                  'Description' => $value_rules['description'] ?? NULL,
                ],
              ],
            ];
          }

          $ip_permission_cnt++;
        }
      }

      $results['SecurityGroups'][$key] = [
        'Description'         => $value['description'] ?? NULL,
        'GroupName'           => $value['name'] ?? NULL,
        'IpPermissions'       => $ip_permissions ?? [],
        'OwnerId'             => $value['tenant_id'] ?? NULL,
        'GroupId'             => $value['id'] ?? NULL,
        'IpPermissionsEgress' => $ip_permissions_egress ?? [],
      ];
    }

    return $results;
  }

  /**
   * Create security group list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The security group create API response.
   */
  public function createSecurityGroup(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['security_group'] = [
      'name' => $params['GroupName'],
      'description' => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/security-groups?project_id={$project_id}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $security_group = $response['body']['security_group'] ?? [];

    return [
      'Description'         => $security_group['description'] ?? NULL,
      'GroupName'           => $security_group['name'] ?? NULL,
      'IpPermissions'       => [],
      'OwnerId'             => $security_group['tenant_id'] ?? NULL,
      'GroupId'             => $security_group['id'] ?? NULL,
      'IpPermissionsEgress' => [],
    ];
  }

  /**
   * Update security group List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The security group update API response.
   */
  public function updateSecurityGroup(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['security_group'] = [
      'name' => $params['GroupName'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/security-groups/{$params['GroupId']}?project_id={$project_id}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $security_group = $response['body']['security_group'] ?? [];

    return [
      'Description'         => $security_group['description'] ?? NULL,
      'GroupName'           => $security_group['name'] ?? NULL,
      'IpPermissions'       => [],
      'OwnerId'             => $security_group['tenant_id'] ?? NULL,
      'GroupId'             => $security_group['id'] ?? NULL,
      'IpPermissionsEgress' => [],
    ];
  }

  /**
   * Delete security group List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function deleteSecurityGroup(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/security-groups/{$params['GroupId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * Create security group Rule.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The security group rule create API response.
   */
  public function createSecurityGroupRule(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $cidr_ip = !empty($params['CidrIp']) ? $params['CidrIp'] : NULL;
    $remote_ip = !empty($params['CidrIpv6']) ? $params['CidrIpv6'] : $cidr_ip;
    $remote_group_id = !empty($params['RemoteGroupId']) ? $params['RemoteGroupId'] : NULL;

    $body['security_group_rule'] = [
      'direction'         => $params['Direction'],
      'port_range_min'    => $params['FromPort'],
      'port_range_max'    => $params['ToPort'],
      'ethertype'         => $params['Source'],
      'protocol'          => $params['IpProtocol'],
      'remote_group_id'   => $remote_group_id,
      'security_group_id' => $params['GroupId'],
      'remote_ip_prefix'  => $remote_ip,
      'description'       => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/security-group-rules?project_id={$project_id}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $security_group_rule = $response['body']['security_group_rule'] ?? [];

    $ip_permissions = $ip_permissions_egress = [];

    if (!empty($security_group_rule['direction'])
      && $security_group_rule['direction'] === 'egress') {
      $ip_permissions_egress = [
        'RuleId'     => $security_group_rule['id'] ?? NULL,
        'FromPort'   => $security_group_rule['port_range_min'] ?? 1,
        'IpProtocol' => $security_group_rule['protocol'] ?? '-1',
        'ToPort'     => $security_group_rule['port_range_max'] ?? 65535,
      ];

      if (!empty($security_group_rule['remote_group_id'])) {
        $ip_permissions_egress += [
          'UserIdGroupPairs' => [
            [
              'GroupId' => $security_group_rule['remote_group_id'] ?? NULL,
              'UserId'  => $security_group_rule['tenant_id'] ?? NULL,
            ],
          ],
        ];
      }

      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv6')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions_egress += [
          'Ipv6Ranges' => [
            [
              'CidrIpv6'    => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv4')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions_egress += [
          'IpRanges' => [
            [
              'CidrIp'      => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
    }
    else {
      $ip_permissions = [
        'RuleId'     => $security_group_rule['id'] ?? NULL,
        'FromPort'   => $security_group_rule['port_range_min'] ?? 1,
        'IpProtocol' => $security_group_rule['protocol'] ?? '-1',
        'ToPort'     => $security_group_rule['port_range_max'] ?? 65535,
      ];

      if (!empty($security_group_rule['remote_group_id'])) {
        $ip_permissions += [
          'UserIdGroupPairs' => [
            [
              'GroupId' => $security_group_rule['remote_group_id'] ?? NULL,
              'UserId'  => $security_group_rule['tenant_id'] ?? NULL,
            ],
          ],
        ];
      }

      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv6')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions += [
          'Ipv6Ranges' => [
            [
              'CidrIpv6'    => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv4')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions += [
          'IpRanges' => [
            [
              'CidrIp'      => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
    }

    return [
      'RuleId'              => $security_group_rule['id'] ?? NULL,
      'SecurityGroups'      => ['ip_permissions' => $ip_permissions ?? []],
      'IpPermissionsEgress' => $ip_permissions_egress ?? [],
    ];
  }

  /**
   * Revoke security group list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function revokeSecurityGroupRule(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/security-group-rules/{$params['RuleId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * Get Subnet Detail.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Subnet Detail list API response.
   */
  public function describeSubnetDetail(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'get',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/subnets/{$params['SubnetId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['body']['subnet'] ?? [];
  }

  /**
   * Get network interface list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface list API response.
   */
  public function describeNetworkInterfaces(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'get',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/networks",
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $networks = $response['body']['networks'] ?? [];

    $results['NetworkInterfaces'] = [];

    foreach ($networks ?: [] as $key => $value) {
      $subnets = [];
      foreach ($value['subnets'] ?: [] as $subnet_id) {
        $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
        $subnets[] = $subnet_result['name'];

      }
      $subnet_names = implode(', ', $subnets);

      $results['NetworkInterfaces'][$key] = [
        'Name'               => $value['name'] ?? NULL,
        'Description'        => $value['description'] ?? NULL,
        'Groups'             => [],
        'MacAddress'         => NULL,
        'OwnerId'            => $value['tenant_id'] ?? NULL,
        'NetworkInterfaceId' => $value['id'] ?? NULL,
        'PrivateIpAddress'   => NULL,
        'PrivateIpAddresses' => [],
        'RequesterManaged'   => NULL,
        'SourceDestCheck'    => NULL,
        'Status'             => $value['status'] === 'ACTIVE' ? 'available' : '',
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => [],
        'VpcId'              => NULL,
        'External'           => $value['router:external'],
      ];
    }

    return $results;
  }

  /**
   * Create network interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface create API response.
   */
  public function createNetworkInterface(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['network'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/networks?project_id={$project_id}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $network_interface = $response['body']['network'] ?? [];

    $subnets = [];
    foreach ($network_interface['subnets'] ?: [] as $subnet_id) {
      $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
      $subnets[] = $subnet_result['name'];

    }
    $subnet_names = implode(', ', $subnets);

    return [
      'NetworkInterface' => [
        'Name'               => $network_interface['name'] ?? NULL,
        'Description'        => $network_interface['description'] ?? NULL,
        'Groups'             => [],
        'MacAddress'         => NULL,
        'OwnerId'            => $network_interface['tenant_id'] ?? NULL,
        'NetworkInterfaceId' => $network_interface['id'] ?? NULL,
        'PrivateIpAddress'   => NULL,
        'PrivateIpAddresses' => [],
        'RequesterManaged'   => NULL,
        'SourceDestCheck'    => NULL,
        'Status'             => $network_interface['status'] ?? NULL,
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => [],
        'VpcId'              => NULL,
      ],
    ];
  }

  /**
   * Update network interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface update API response.
   */
  public function updateNetworkInterface(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['network'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/networks/{$params['NetworkInterfaceId']}?project_id={$project_id}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $network_interface = $response['body']['network'] ?? [];

    $subnets = [];
    foreach ($network_interface['subnets'] ?: [] as $subnet_id) {
      $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
      $subnets[] = $subnet_result['name'];

    }
    $subnet_names = implode(', ', $subnets);

    return [
      'NetworkInterface' => [
        'Name'               => $network_interface['name'] ?? NULL,
        'Description'        => $network_interface['description'] ?? NULL,
        'Groups'             => [],
        'MacAddress'         => NULL,
        'OwnerId'            => $network_interface['tenant_id'] ?? NULL,
        'NetworkInterfaceId' => $network_interface['id'] ?? NULL,
        'PrivateIpAddress'   => NULL,
        'PrivateIpAddresses' => [],
        'RequesterManaged'   => NULL,
        'SourceDestCheck'    => NULL,
        'Status'             => $network_interface['status'] ?? NULL,
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => [],
        'VpcId'              => NULL,
      ],
    ];
  }

  /**
   * Delete network interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return bool
   *   Succeeded or failed.
   */
  public function deleteNetworkInterface(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/networks/{$params['NetworkInterfaceId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * Get Floating IP List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP list API response.
   */
  public function describeAddresses(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'get',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/floatingips?project_id={$project_id}",
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $floating_ips = $response['body']['floatingips'] ?? [];

    $results['Addresses'] = [];

    foreach ($floating_ips ?: [] as $key => $value) {
      $results['Addresses'][$key] = [
        'FloatingIpId'            => $value['id'],
        'PublicIp'                => $value['floating_ip_address'] ?? NULL,
        'PrivateIpAddress'        => $value['fixed_ip_address'] ?? NULL,
        'NetworkInterfaceOwnerId' => $value['tenant_id'] ?? NULL,
        'NetworkInterfaceId'      => $value['port_details'] ? $value['port_details']['network_id'] : NULL,
        'InstanceId'              => $value['port_details'] ? $value['port_details']['device_id'] : NULL,
        'AssociationId'           => $value['port_id'] ?? NULL,
        'AllocationId'            => $value['floating_network_id'] ?? NULL,
      ];
    }
    return $results;
  }

  /**
   * Create Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP create API response.
   */
  public function allocateAddress(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['floatingip'] = [
      'floating_network_id' => $params['AllocationId'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/floatingips",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $floating_ip = $response['body']['floatingip'] ?? [];

    return [
      'FloatingIpId'            => $floating_ip['id'] ?? NULL,
      'PublicIp'                => $floating_ip['floating_ip_address'] ?? NULL,
      'PrivateIpAddress'        => $floating_ip['fixed_ip_address'] ?? NULL,
      'NetworkInterfaceOwnerId' => $floating_ip['tenant_id'] ?? NULL,
      'NetworkInterfaceId'      => $floating_ip['port_details'] ? $floating_ip['port_details']['network_id'] : NULL,
      'InstanceId'              => $floating_ip['port_details'] ? $floating_ip['port_details']['device_id'] : NULL,
      'Domain'                  => $floating_ip['dns_domain'] ?? NULL,
      'AssociationId'           => $floating_ip['port_id'] ?? NULL,
      'AllocationId'            => $floating_ip['floating_network_id'] ?? NULL,
    ];
  }

  /**
   * Associate Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP associate API response.
   */
  public function associateAddress(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['floatingip'] = [
      'port_id' => $params['PortId'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/floatingips/{$params['FloatingIpId']}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $floating_ip = $response['body']['floatingip'] ?? [];

    return !empty($response) && !empty($response['body'])
      && $response['status_code'] === 200 ?
      [
        'FloatingIpId'            => $floating_ip['id'] ?? NULL,
        'PublicIp'                => $floating_ip['floating_ip_address'] ?? NULL,
        'PrivateIpAddress'        => $floating_ip['fixed_ip_address'] ?? NULL,
        'NetworkInterfaceOwnerId' => $floating_ip['tenant_id'] ?? NULL,
        'NetworkInterfaceId'      => !empty($floating_ip['port_details']) ? $floating_ip['port_details']['network_id'] : NULL,
        'InstanceId'              => !empty($floating_ip['port_details']) ? $floating_ip['port_details']['device_id'] : NULL,
        'Domain'                  => $floating_ip['dns_domain'] ?? NULL,
        'AssociationId'           => $floating_ip['port_id'] ?? NULL,
      ]
    : [];
  }

  /**
   * Disassociate Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP associate API response.
   */
  public function disassociateAddress(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $body['floatingip'] = [
      'port_id' => NULL,
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/floatingips/{$params['FloatingIpId']}",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $floating_ip = $response['body']['floatingip'] ?? [];

    return [
      'FloatingIpId'            => $floating_ip['id'] ?? NULL,
      'PublicIp'                => $floating_ip['floating_ip_address'] ?? NULL,
      'PrivateIpAddress'        => $floating_ip['fixed_ip_address'] ?? NULL,
      'NetworkInterfaceOwnerId' => $floating_ip['tenant_id'] ?? NULL,
      'NetworkInterfaceId'      => $floating_ip['port_details'] ? $floating_ip['port_details']['network_id'] : NULL,
      'InstanceId'              => $floating_ip['port_details'] ? $floating_ip['port_details']['device_id'] : NULL,
      'Domain'                  => $floating_ip['dns_domain'] ?? NULL,
      'AssociationId'           => $floating_ip['port_id'] ?? NULL,
    ];
  }

  /**
   * Instance Port Interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance Port Interface list API response.
   */
  public function instancePortInterface(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? NULL;

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/compute/{$compute_version}/servers/{$params['InstanceId']}/os-interface",
    ];

    $response = $this->getClient($credentials);

    return $response['body']['interfaceAttachments'] ?? [];
  }

  /**
   * Get NetworkInterface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The NetworkInterface list API response.
   */
  public function describeNetworkInterfacesPorts(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'get',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/ports?project_id={$project_id}",
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $ports = $response['body']['ports'] ?? [];

    $results = [];

    foreach ($ports ?: [] as $port) {
      $results[$port['network_id']][$port['id']] = $port['id'];
    }

    return $results;
  }

  /**
   * Delete Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP delete API response.
   */
  public function releaseAddress(array $params = []) {
    $this->loadCredentials();
    $api_endpoint_port = $this->credentials['api_endpoint_port'] ?? NULL;
    $network_version = $this->credentials['network_version'] ?? NULL;

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => ":{$api_endpoint_port}/{$network_version}/floatingips/{$params['FloatingIpId']}",
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * {@inheritdoc}
   */
  public function describeAvailabilityZones(array $params = []) {
    $this->loadCredentials();
    $compute_version = $this->credentials['compute_version'] ?? NULL;

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/compute/{$compute_version}/os-availability-zone",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    $availability_zones = $response['body']['availabilityZoneInfo'] ?? [];

    $results['AvailabilityZones'] = [];

    foreach ($availability_zones ?: [] as $key => $value) {
      $results['AvailabilityZones'][$key] = [
        'ZoneName' => $value['zoneName'] ?? NULL,
        'State'   => $value['zoneState']['available'] === TRUE ? 'available' : '',
      ];
    }
    return $results;
  }

  /**
   * Call the API for updated entities and store them as Instance entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateInstanceEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = ''): bool {
    $updated = FALSE;
    // Call the api and get all instances.
    $result = $this->describeInstances($params);
    if ($result !== NULL) {
      $all_instances = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the instances by setting up
      // the array with the instance_id.
      foreach ($all_instances ?: [] as $instance) {
        $stale[$instance->getInstanceId()] = $instance;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Instance Update');
      // Loop through the reservations and store each one as an Instance entity.
      // Testcase throws an undefined index: Reservations error to resolve
      // that error needs to be add isset condition.
      if (isset($result['Reservations'])) {
        foreach ($result['Reservations'] ?: [] as $reservation) {

          foreach ($reservation['Instances'] ?: [] as $instance) {
            // Keep track of instances that do not exist anymore
            // delete them after saving the rest of the instances.
            if (!empty($stale[$instance['InstanceId']])) {
              unset($stale[$instance['InstanceId']]);
            }
            // Store the Reservation OwnerId in instance so batch
            // callback has access.
            $instance['reservation_ownerid'] = $reservation['OwnerId'] ?? NULL;
            $instance['reservation_id'] = $reservation['ReservationId'] ?? NULL;

            $batch_builder->addOperation([
              OpenStackBatchOperations::class,
              'updateInstance',
            ], [$cloud_context, $instance]);

          }
        }
      }
      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Update the Instances.
   *
   * @param array $params
   *   Parameters array to send to API.
   * @param bool $clear
   *   TRUE to clear stale instances.
   *
   * @return bool
   *   Indicates success or failure.
   */
  public function updateInstances(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_instance';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateInstanceEntities($params, $clear, $entity_type, $this->cloudContext);
    $this->lock->release($lock_name);

    return $updated;
  }

  /**
   * Update the OpenStackImages.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateImages(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_image';
    $lock_name = $this->getLockKey($entity_type);
    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateImageEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as Image entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateImageEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = ''): bool {
    $updated = FALSE;

    // Load all entities by $this->cloudContext.
    $image_entities = $this->entityTypeManager->getStorage($entity_type)->loadByProperties(
      ['cloud_context' => $this->cloudContext]
    );

    $result = $this->describeImages($params);

    // Keep isset(), and do not replace !empty() here. We need to check for
    // instance testcase notice because the instance testcase throws an error
    // of undefined index notice.
    if (isset($result)) {

      $stale = [];
      // Make it easier to lookup the images by setting up
      // the array with the image_id.
      foreach ($image_entities ?: [] as $image) {
        $stale[$image->getImageId()] = $image;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Image Update');

      if (!empty($result['Images'])) {
        foreach ($result['Images'] ?: [] as $image) {
          // Keep track of images that do not exist anymore
          // delete them after saving the rest of the images.
          if (!empty($stale[$image['ImageId']])) {
            unset($stale[$image['ImageId']]);
          }

          $batch_builder->addOperation([
            OpenStackBatchOperations::class,
            'updateImage',
          ], [$cloud_context, $image]);
        }
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as key pair entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param array $params
   *   Key Name info.
   * @param bool $update
   *   TRUE to update Key info.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateKeyPairEntities($entity_type = '', $cloud_context = '', array $params = [], $update = TRUE): bool {
    $updated = FALSE;
    $result = $this->describeKeyPairs();

    if (!empty($result)) {

      $all_keys = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_keys ?: [] as $key) {
        $stale[$key->getKeyPairName()] = $key;
      }
      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Keypair Update');

      foreach ($result['KeyPairs'] ?: [] as $key_pair) {
        // Keep track of key pair that do not exist anymore
        // delete them after saving the rest of the key pair.
        if (!empty($stale[$key_pair['KeyName']])) {
          unset($stale[$key_pair['KeyName']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateKeyPair',
        ], [$cloud_context, $key_pair]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * Update the OpenStack key pairs.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateKeyPairs(): bool {
    $entity_type = 'openstack_key_pair';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateKeyPairEntities($entity_type, $this->cloudContext);
    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as volume entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param array $params
   *   Volume info.
   * @param bool $update
   *   TRUE to update volume.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateVolumeEntities($entity_type = '', $cloud_context = '', array $params = [], $update = TRUE): bool {
    $updated = FALSE;
    $result = $this->describeVolumes();

    $properties = empty($params['VolumeIds'][0])
      ? ['cloud_context' => $cloud_context]
      : [
        'cloud_context' => $cloud_context,
        'volume_id' => $params['VolumeIds'][0],
      ];
    $all_volumes = $this->loadAllEntities($entity_type, $properties);
    $stale = [];
    // Make it easier to lookup the groups by setting up
    // the array with the group_id.
    foreach ($all_volumes ?: [] as $volume) {
      $stale[$volume->getVolumeId()] = $volume;
    }
    $snapshot_id_name_map = $this->getSnapshotIdNameMap(!empty($result['Volumes']) ? $result['Volumes'] : []);

    /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
    $batch_builder = $this->initBatch('Volume Update');

    if (!empty($result)) {
      foreach ($result['Volumes'] ?: [] as $volume) {
        // Keep track of network interfaces that do not exist anymore
        // delete them after saving the rest of the network interfaces.
        if (!empty($stale[$volume['VolumeId']])) {
          unset($stale[$volume['VolumeId']]);
        }
        if ($update) {
          $batch_builder->addOperation([
            OpenStackBatchOperations::class,
            'updateVolume',
          ], [$cloud_context, $volume, $snapshot_id_name_map]);
        }
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = empty($params['VolumeIds'][0]) ? TRUE : !empty($result['Volumes']);
    }
    elseif (empty($result) && !empty($params['VolumeIds'][0])) {
      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
    }
    return $updated;
  }

  /**
   * Update the volumes.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateVolumes(): bool {
    $entity_type = 'openstack_volume';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateVolumeEntities($entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Get snapshot id and name.
   *
   * @param array $volumes
   *   The volumes array.
   *
   * @return array
   *   Return map array.
   */
  public function getSnapshotIdNameMap(array $volumes = []) {
    $snapshot_ids = array_filter(array_column($volumes, 'SnapshotId'));
    if (empty($snapshot_ids)) {
      return [];
    }

    $map = [];
    foreach ($snapshot_ids ?: [] as $snapshot_id) {
      $map[$snapshot_id] = $snapshot_id;
    }

    $result = $this->describeSnapshots();

    if (!empty($result)) {
      foreach ($result['Snapshots'] ?: [] as $snapshot) {
        $snapshot_id = $snapshot['SnapshotId'];
        if (!array_key_exists($snapshot_id, $map)) {
          continue;
        }

        $map[$snapshot_id] = $this->getTagName($snapshot, '');
      }
    }

    return $map;
  }

  /**
   * Call the API for updated entities and store them as snapshot entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param array $params
   *   Snapshot IDs.
   * @param bool $update
   *   TRUE to update info.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateSnapshotEntities($entity_type = '', $cloud_context = '', array $params = [], $update = TRUE): bool {
    $updated = FALSE;

    $result = $this->describeSnapshots();

    if (!empty($result)) {

      $all_snapshots = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the snapshot by setting up
      // the array with the snapshot_id.
      foreach ($all_snapshots ?: [] as $snapshot) {
        $stale[$snapshot->getSnapshotId()] = $snapshot;
      }
      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Snapshot Update');

      foreach ($result['Snapshots'] ?: [] as $snapshot) {
        // Keep track of snapshot that do not exist anymore
        // delete them after saving the rest of the snapshots.
        if (!empty($stale[$snapshot['SnapshotId']])) {
          unset($stale[$snapshot['SnapshotId']]);
        }

        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateSnapshot',
        ], [$cloud_context, $snapshot]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * Update the snapshots.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateSnapshots(): bool {
    $entity_type = 'openstack_snapshot';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateSnapshotEntities($entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Update the OpenStackSecurityGroups.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateSecurityGroups(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_security_group';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateSecurityGroupEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as SecurityGroup entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param bool $update
   *   TRUE to update SG info.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateSecurityGroupEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '', $update = TRUE): bool {

    $cloud_config = $this->entityTypeManager
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'openstack',
        'cloud_context' => $cloud_context,
      ]);

    $cloud_config = array_shift($cloud_config);

    $params['project_id'] = $cloud_config->get('field_project_id')->value ?? NULL;

    $updated = FALSE;
    $result = $this->describeSecurityGroups($params);

    if (!empty($result)) {

      $all_groups = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_groups ?: [] as $group) {
        $stale[$group->getGroupId()] = $group;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Security Group Update');

      foreach ($result['SecurityGroups'] ?: [] as $security_group) {

        // Keep track of instances that do not exist anymore
        // delete them after saving the rest of the instances.
        if (!empty($stale[$security_group['GroupId']])) {
          unset($stale[$security_group['GroupId']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateSecurityGroup',
        ], [$cloud_context, $security_group]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * Update the OpenStackNetworkInterfaces.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateNetworkInterfaces(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_network_interface';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateNetworkInterfaceEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call API for updated entities and store them as NetworkInterface entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param bool $update
   *   TRUE to update info.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateNetworkInterfaceEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '', $update = TRUE): bool {

    $cloud_config = $this->entityTypeManager
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'openstack',
        'cloud_context' => $cloud_context,
      ]);

    $cloud_config = array_shift($cloud_config);

    $params['project_id'] = $cloud_config->get('field_project_id')->value ?? NULL;

    $updated = FALSE;
    $result = $this->describeNetworkInterfaces($params);

    if (!empty($result)) {

      $all_interfaces = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_interfaces ?: [] as $interface) {
        $stale[$interface->getNetworkInterfaceId()] = $interface;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Network Interface Update');

      foreach ($result['NetworkInterfaces'] ?: [] as $network_interface) {
        // Keep track of network interfaces that do not exist anymore
        // delete them after saving the rest of the network interfaces.
        if (!empty($stale[$network_interface['NetworkInterfaceId']])) {
          unset($stale[$network_interface['NetworkInterfaceId']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateNetworkInterface',
        ], [$cloud_context, $network_interface]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Update the OpenStackFloatingIps.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateFloatingIps(): bool {
    $entity_type = 'openstack_floating_ip';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateFloatingIpEntities($entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as FloatingIp entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateFloatingIpEntities($entity_type = '', $cloud_context = ''): bool {
    $updated = FALSE;
    $result = $this->describeAddresses();

    if (!empty($result)) {

      $all_ips = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_ips ?: [] as $ip) {
        $stale[$ip->getPublicIp()] = $ip;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('FloatingIp Update');

      foreach ($result['Addresses'] ?: [] as $floating_ip) {
        // Keep track of IPs that do not exist anymore
        // delete them after saving the rest of the IPs.
        if (!empty($stale[$floating_ip['PublicIp']])) {
          unset($stale[$floating_ip['PublicIp']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateFloatingIp',
        ], [$cloud_context, $floating_ip]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailabilityZones($with_id = FALSE, array $credentials = []): array {
    $zones = [];
    $result = $this->describeAvailabilityZones([], $credentials);
    if (!empty($result)) {
      foreach ($result['AvailabilityZones'] ?: [] as $zone_info) {
        if ($with_id) {
          $name = sprintf('%s (%s)', $zone_info['ZoneName'], $zone_info['ZoneId']);
        }
        else {
          $name = $zone_info['ZoneName'];
        }
        $zones[$zone_info['ZoneName']] = $name;
      }
    }
    return $zones;
  }

  /**
   * Helper method to do a Guzzle HTTP Request.
   *
   * @param string $http_method
   *   The http method in lower case.
   * @param string $endpoint
   *   The endpoint to access.
   * @param array $params
   *   Parameters to pass to the request.
   *
   * @return array
   *   Array interpretation of the response body.
   */
  private function request($http_method, $endpoint, array $params = []) {
    $output = [];

    try {
      $response = $this->httpClient->$http_method(
        rtrim($this->credentials['authUrl'], '/') . $endpoint,
        $params
      );

      if (!empty($response)) {
        $status_code = $response->getStatusCode();
        $headers = $response->getHeaders();
        $output['auth_token'] = $headers['X-Subject-Token'][0] ?? '';
        $output['status_code'] = $status_code;
        $body = $response->getBody()->getContents();
      }
    }
    catch (GuzzleException $error) {
      $response = $error->getResponse();
      $error_content = [];
      $content = '';
      if (!empty($response)) {
        $content = $response->getBody()->getContents();
        $status_code = $response->getStatusCode();
        $output['status_code'] = $status_code;
        $error_content = json_decode($content, TRUE);
      }

      $error_content = !empty($error_content) ? reset($error_content) : [];

      if (!empty($error_content)) {
        $this->messenger->addError($this->t('Status Code: @status_code', [
          '@status_code' => $error_content['code'] ?? '',
        ]));

        $this->messenger->addError($this->t('Error Message: @msg', [
          '@msg' => $error_content['message'],
        ]));
      }
      else {
        $this->messenger->addError($this->t('OpenStack API error. Error details are as follows: @response', [
          '@response' => $error->getMessage(),
        ]));
      }

      $message = new FormattableMarkup(
        'OpenStack API error. Error details are as follows:<pre>@response</pre>', [
          '@response' => !empty($content)
          ? $content
          : $error->getMessage(),
        ]
      );
      $this->logError('Remote API Connection', $error, $message, FALSE);
    }
    catch (\Exception $error) {
      $this->messenger->addError($this->t('An unknown error occurred while
      trying to connect to the remote API. This is not a Guzzle error, not an
      error in the remote API, rather a generic local error occurred.
      The reported error was @error', [
        '@error' => $error->getMessage(),
      ]));

      $this->logError('Remote API Connection', $error, $this->t('An unknown error
      occurred while trying to connect to the remote API. This is not a Guzzle
      error, nor an error in the remote API, rather a generic local error
      occurred. The reported error was @error',
        ['@error' => $error->getMessage()]
      ), FALSE);
    }
    $body = empty($body) ? [] : json_decode($body, TRUE);
    $output['body'] = $body;
    return $output;
  }

  /**
   * Generate a lock key based on entity name.
   *
   * @param string $name
   *   The entity name.
   *
   * @return string
   *   The lock key.
   */
  protected function getLockKey($name) {
    return $this->cloudContext . '_' . $name;
  }

  /**
   * Initialize a new batch builder.
   *
   * @param string $batch_name
   *   The batch name.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The initialized batch object.
   */
  protected function initBatch($batch_name) {
    return (new BatchBuilder())
      ->setTitle($batch_name);
  }

  /**
   * Run the batch job to process entities.
   *
   * @param \Drupal\Core\Batch\BatchBuilder $batch_builder
   *   The batch builder object.
   */
  protected function runBatch(BatchBuilder $batch_builder) {
    try {
      // Log the start time.
      $start = $this->getTimestamp();
      $batch_array = $batch_builder->toArray();
      batch_set($batch_array);
      // Reset the progressive so batch works with out a web head.
      $batch = &batch_get();
      $batch['progressive'] = FALSE;
      // Check if using Drush to process the queue.
      // using drush_backend_batch_process() will fix the "Route not found"
      // error.
      if (PHP_SAPI === 'cli' && function_exists('drush_backend_batch_process')) {
        drush_backend_batch_process();
      }
      else {
        batch_process();
      }
      // Log the end time.
      $end = $this->getTimestamp();
      $this->logger('openstack_rest')->info($this->t('@updater - @cloud_context: Batch operation took @time seconds.', [
        '@cloud_context' => $this->cloudContext,
        '@updater' => $batch_array['title'],
        '@time' => $end - $start,
      ]));
    }
    catch (Exception $e) {
      $this->handleException($e);
    }
    finally {
      // Reset the batch otherwise this operation hangs when using Drush.
      // https://www.drupal.org/project/drupal/issues/3166042
      $batch = [];
    }
  }

  /**
   * Log errors to watchdog and throw an exception.
   *
   * @param string $type
   *   The error type.
   * @param \Exception $exception
   *   The exception object.
   * @param string $message
   *   The message to log.
   * @param bool $throw
   *   TRUE to throw exception.
   *
   * @throws \Drupal\openstack\Service\Rest\OpenStackServiceException
   */
  private function logError($type, \Exception $exception, $message, $throw = TRUE) {
    watchdog_exception($type, $exception, $message);
    if ($throw === TRUE) {
      throw new OpenStackServiceException($message);
    }
  }

  /**
   * Helper static method to clear cache.
   */
  public static function clearCacheValue() {
    \Drupal::cache('menu')->invalidateAll();
    \Drupal::service('cache.render')->deleteAll();
    \Drupal::service('router.builder')->rebuild();
    \Drupal::service('plugin.cache_clearer')->clearCachedDefinitions();
  }

  /**
   * Helper method to get the current timestamp.
   *
   * @return int
   *   The current timestamp.
   */
  protected function getTimestamp() {
    return time();
  }

  /**
   * Get Tag name.
   *
   * @param array $openstack_obj
   *   The openstack object.
   * @param string $default_value
   *   The default value.
   *
   * @return string
   *   Return name of corresponding resource.
   */
  public function getTagName(array $openstack_obj, $default_value) {
    $name = $default_value;
    if (empty($openstack_obj['name'])) {
      return $name;
    }

    if (!empty($openstack_obj['name'])) {
      return $openstack_obj['name'];
    }

    foreach ($openstack_obj['Tags'] ?: [] as $tag) {
      if ($tag['Key'] === 'Name' && !empty($tag['Value'])) {
        $name = $tag['Value'];
        break;
      }
    }
    return $name;
  }

  /**
   * Setup IP Permissions.
   *
   * @param object $security_group
   *   The security group entity.
   * @param string $field
   *   Field to used for lookup.
   * @param array $ec2_permissions
   *   Permissions array from Ec2.
   */
  public function setupIpPermissions(&$security_group, $field, array $ec2_permissions) {
    // Permissions are always overwritten with the latest from
    // EC2.  The reason is that there is no way to guarantee a 1
    // to 1 mapping from the $security_group['IpPermissions'] array.
    // There is no IP permission ID coming back from EC2.
    // Clear out all items before re-adding them.
    $i = $security_group->$field->count() - 1;

    while ($i >= 0) {
      if ($security_group->$field->get($i)) {
        $security_group->$field->removeItem($i);
      }
      $i--;
    }

    // Setup all permission objects.
    $count = 0;
    foreach ($ec2_permissions ?: [] as $permissions) {
      $permission_objects = $this->setupIpPermissionObject($permissions);
      // Loop through the permission objects and add them to the
      // security group ip_permission field.
      foreach ($permission_objects ?: [] as $permission) {
        $security_group->$field->set($count, $permission);
        $count++;
      }
    }
  }

  /**
   * Setup the ip_permission field given the inbound security group array.
   *
   * The array comes from DescribeSecurityGroup Amazon EC2 API call.
   *
   * @param array $ec2_permission
   *   An array object of EC2permission.
   *
   * @return array
   *   An array of \Drupal\Core\Field\FieldItemInterface.
   */
  public function setupIpPermissionObject(array $ec2_permission) {
    $ip_permissions = [];

    // $file_path = \Drupal::service('file_system')->realpath("public://");
    // $handle = fopen($file_path . "/myfile.txt", "a+");
    // fwrite($handle, json_encode($ec2_permission));
    // fclose($handle);
    // Get the field definition for an IpPermission object.
    $definition = $this->entityFieldManager->getBaseFieldDefinitions('openstack_security_group');

    /* Setup the more global attributes.
     * NOTE: DO NOT change from isset() to !empty() here since
     * !empty($ec2_permission['FromPort']) is considered empty
     * in PHP when $ec2_permission['FromPort'] is a zero (0) value.
     */
    $rule_id = $ec2_permission['RuleId'] ?? NULL;
    $from_port = (string) ($ec2_permission['FromPort']);
    $to_port = (string) ($ec2_permission['ToPort']);

    $ip_protocol = $ec2_permission['IpProtocol'];

    // To keep things consistent, if ip_protocol is -1,
    // set from_port and to_port as 0-65535.
    if ($ip_protocol === '-1') {
      $from_port = '0';
      $to_port = '65535';
    }

    if (!empty($ec2_permission['IpRanges'])) {
      // Create a IPv4 permission object.
      foreach ($ec2_permission['IpRanges'] ?: [] as $ip_range) {
        $ip_range_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $ip_range_permission->source = 'ip4';
        $ip_range_permission->cidr_ip = $ip_range['CidrIp'];
        $ip_range_permission->description = $ip_range['Description'] ?? NULL;

        $ip_range_permission->rule_id = $rule_id;
        $ip_range_permission->from_port = $from_port;
        $ip_range_permission->to_port = $to_port;
        $ip_range_permission->ip_protocol = $ip_protocol;

        $ip_permissions[] = $ip_range_permission;
      }
    }

    if (!empty($ec2_permission['Ipv6Ranges'])) {
      // Create IPv6 permissions object.
      foreach ($ec2_permission['Ipv6Ranges'] ?: [] as $ip_range) {
        $ip_v6_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $ip_v6_permission->rule_id = $rule_id;
        $ip_v6_permission->source = 'ip6';
        $ip_v6_permission->cidr_ip_v6 = $ip_range['CidrIpv6'];
        $ip_v6_permission->description = $ip_range['Description'] ?? NULL;
        $ip_v6_permission->from_port = $from_port;
        $ip_v6_permission->to_port = $to_port;
        $ip_v6_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $ip_v6_permission;
      }
    }

    if (!empty($ec2_permission['UserIdGroupPairs'])) {
      // Create Group permissions object.
      foreach ($ec2_permission['UserIdGroupPairs'] ?: [] as $group) {
        $group_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $group_permission->rule_id = $rule_id;
        $group_permission->source = 'group';
        $group_permission->group_id = $group['GroupId'] ?? NULL;
        $group_permission->group_name = $group['GroupName'] ?? NULL;
        $group_permission->user_id = $group['UserId'] ?? NULL;
        $group_permission->peering_status = $group['PeeringStatus'] ?? NULL;
        $group_permission->vpc_id = $group['VpcId'] ?? NULL;
        $group_permission->peering_connection_id = $group['VpcPeeringConnectionId'] ?? NULL;
        $group_permission->description = $group['Description'] ?? NULL;
        $group_permission->from_port = $from_port;
        $group_permission->to_port = $to_port;
        $group_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $group_permission;
      }
    }

    if (!empty($ec2_permission['PrefixListIds'])) {
      foreach ($ec2_permission['PrefixListIds'] ?: [] as $prefix) {
        $prefix_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $prefix_permission->rule_id = $rule_id;
        $prefix_permission->source = 'prefix';
        $prefix_permission->prefix_list_id = $prefix['PrefixListId'];
        $prefix_permission->description = $prefix['Description'] ?? NULL;
        $prefix_permission->from_port = $from_port;
        $prefix_permission->to_port = $to_port;
        $prefix_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $prefix_permission;
      }
    }

    return $ip_permissions;
  }

  /**
   * Method to clear all entities out of the system.
   */
  public function clearAllEntities() {
    $timestamp = $this->getTimestamp();
    $this->clearEntities('openstack_instance', $timestamp);
    $this->clearEntities('openstack_image', $timestamp);
    $this->clearEntities('openstack_key_pair', $timestamp);
    $this->clearEntities('openstack_volume', $timestamp);
    $this->clearEntities('openstack_snapshot', $timestamp);
    $this->clearEntities('openstack_security_group', $timestamp);
    $this->clearEntities('openstack_network_interface', $timestamp);
    $this->clearEntities('openstack_floating_ip', $timestamp);
  }

  /**
   * {@inheritdoc}
   */
  public function createResourceQueueItems(CloudConfigInterface $cloud_config): void {
    $queue_limit = $this->configFactory->get('openstack.settings')->get('openstack_queue_limit');
    $update_resources_queue = $this->queueFactory->get('openstack_update_resources_queue');
    if ($update_resources_queue->numberOfItems() >= $queue_limit) {
      return;
    }
    $method_names = [
      'updateInstances',
      'updateImages',
      'updateKeyPairs',
      'updateVolumes',
      'updateSnapshots',
      'updateSecurityGroups',
      'updateFloatingIps',
      'updateNetworkInterfaces',
    ];
    foreach ($method_names as $method_name) {
      $update_resources_queue->createItem([
        'cloud_context' => $this->cloudContext,
        'openstack_method_name' => $method_name,
      ]);
    }
  }

  /**
   * Get UID tag value.
   *
   * @param array $tags_array
   *   The tags array of API.
   * @param string $entity_type_id
   *   The entity type id.
   *
   * @return int
   *   Return user id.
   */
  public function getUidTagValue(array $tags_array, $entity_type_id): int {
    // @todo Support AwsCloudEntity::getTagCreatedByUid() in OpenStack module.
    $key = $entity_type_id . '_';
    $key .= strpos($entity_type_id, 'aws_cloud') !== FALSE
      ? AwsCloudEntity::getTagCreatedByUid($this->cloudContext)
      : AwsCloudEntityInterface::TAG_CREATED_BY_UID;
    $uid = 0;
    if (!empty($tags_array['Tags'])) {
      foreach ($tags_array['Tags'] ?: [] as $tag) {
        if ($tag['Key'] === $key) {
          $uid = $tag['Value'];
          break;
        }
      }
    }
    return $uid;
  }

  /**
   * Get InstanceUID.
   *
   * @param string $instance_id
   *   The instance id.
   *
   * @return int
   *   Return user id.
   */
  public function getInstanceUid($instance_id) {
    $uid = 0;
    $instance = $this->entityTypeManager
      ->getStorage('openstack_instance')
      ->loadByProperties([
        'instance_id' => $instance_id,
      ]);

    if (count($instance) > 0) {
      $instance = array_shift($instance);
      $uid = $instance->getOwnerId();
    }
    return $uid;
  }

}
